"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from collections.abc import Iterable
from math import cos, pi
from functools import reduce

from enum import Enum
import numpy as np


def enum_missing(cls, value):
    """
    This method is called when `value` is not found in the Enum. This could be the
    case when the input is not lower case but its lower case version is in the Enum.
    Therefore `value` is first converted to lower-case and then compared to the values
    in the Enum.

    If no match is found, an exception is raised. Also note that the exception message
    is more informative than the default one since we here include the allowed values.
    """
    # Check the input in a case insensitive way
    value = value.lower()
    for member in cls:
        if member.value == value:
            return member

    # If no case insensitive match is found, raise an exception
    allowed_values = [member.value for member in cls]
    raise ValueError(
        f"`{value}` is an invalid input for enum `{cls.__name__}`. Only the following "
        f"values are allowed as an input: {', '.join(allowed_values)}."
    )


class SingleQubitPauliEigenstate(str, Enum):
    """
    Supported states in reset operations.
    """

    ZERO = "0"
    ONE = "1"
    PLUS = "+"
    MINUS = "-"
    PLUS_I = "+i"
    MINUS_I = "-i"

    @property
    def pauli_basis(self) -> str:
        """
        Get the Pauli basis of the state.
        """
        if self in (SingleQubitPauliEigenstate.ZERO, SingleQubitPauliEigenstate.ONE):
            return "Z"
        if self in (
            SingleQubitPauliEigenstate.PLUS,
            SingleQubitPauliEigenstate.MINUS,
        ):
            return "X"
        if self in (
            SingleQubitPauliEigenstate.PLUS_I,
            SingleQubitPauliEigenstate.MINUS_I,
        ):
            return "Y"

        raise ValueError(f"Invalid state: {self}")

    @property
    def basis_expectation_value(self) -> int:
        """
        Get the expectation value of the state in the Pauli basis of the state.
        """
        if self in (
            SingleQubitPauliEigenstate.ZERO,
            SingleQubitPauliEigenstate.PLUS,
            SingleQubitPauliEigenstate.PLUS_I,
        ):
            return 1
        if self in (
            SingleQubitPauliEigenstate.ONE,
            SingleQubitPauliEigenstate.MINUS,
            SingleQubitPauliEigenstate.MINUS_I,
        ):
            return -1

        raise ValueError(f"Invalid state: {self}")


class Direction(str, Enum):
    """
    Direction indicator for Operations (e.g. Grow and Shrink).
    """

    LEFT = "left"
    RIGHT = "right"
    TOP = "top"
    BOTTOM = "bottom"

    @classmethod
    def _missing_(cls, value):
        """
        Allow inputs with upper-case characters. For more details, see the
        documentation of `enum_missing` at the beginning of the file.
        """
        return enum_missing(cls, value)

    def __str__(self):
        return str(self.value)

    def to_vector(self) -> tuple[int, int]:
        """
        Convert the direction to a 2D vector.
        """
        if self == Direction.LEFT:
            return (-1, 0)
        if self == Direction.RIGHT:
            return (1, 0)
        if self == Direction.TOP:
            return (0, -1)
        if self == Direction.BOTTOM:
            return (0, 1)

        raise ValueError(f"Direction has no vector representation: {self}")

    @classmethod
    def from_vector(cls, vector: tuple[int, ...], bottom_is_plus=True) -> Direction:
        """
        Get the direction from a 2D vector. The vector should have only
        one non-zero component in the first or the second dimension. The
        direction is determined by the non-zero component. The direction
        is returned as a Direction enum.
        """
        np_vector = np.array(vector)
        if len(np_vector) < 2:
            raise ValueError("Cannot get direction from a 1D vector.")

        if any(np.nonzero(np_vector)[0] > 1):
            raise ValueError(
                "Only the first and the second components of the vector may be non zero"
            )

        # Get the sign of the components in an array
        sign_array = np.sign(np_vector)
        if not bottom_is_plus:
            sign_array[1] *= -1

        # Determine the direction
        match (sign_array[0], sign_array[1]):
            case (1, 0):
                return Direction.RIGHT
            case (-1, 0):
                return Direction.LEFT
            case (0, 1):
                return Direction.BOTTOM
            case (0, -1):
                return Direction.TOP
            case _:
                raise ValueError(
                    "Direction cannot be found from the given vector. The vector should"
                    " have exactly one non-zero component"
                )

    def opposite(self) -> Direction:
        """
        Get the opposite direction.
        """
        if self == Direction.LEFT:
            return Direction.RIGHT
        if self == Direction.RIGHT:
            return Direction.LEFT
        if self == Direction.TOP:
            return Direction.BOTTOM
        if self == Direction.BOTTOM:
            return Direction.TOP

        raise ValueError(f"Direction has no opposite: {self}")

    def to_orientation(self) -> Orientation:
        """Convert the direction to an orientation."""
        return Orientation.from_direction(self)

    def mirror_across_orientation(self, orientation: Orientation) -> Direction:
        """Get the mirrored direction across the given orientation."""
        return self if self.to_orientation() == orientation else self.opposite()


class DiagonalDirection(str, Enum):
    """Diagonal direction indicator for Operations (e.g. Grow and Shrink)."""

    TOP_LEFT = "top-left"
    TOP_RIGHT = "top-right"
    BOTTOM_LEFT = "bottom-left"
    BOTTOM_RIGHT = "bottom-right"

    # Map diagonals to their component directions

    @property
    def components(self):
        """Return the two cardinal directions composing this diagonal."""
        match self.value:
            case "top-left":
                return (Direction.TOP, Direction.LEFT)
            case "top-right":
                return (Direction.TOP, Direction.RIGHT)
            case "bottom-left":
                return (Direction.BOTTOM, Direction.LEFT)
            case "bottom-right":
                return (Direction.BOTTOM, Direction.RIGHT)

    @classmethod
    def _missing_(cls, value):
        """Allow inputs with upper-case characters. For more details, see the
        documentation of `enum_missing` at the beginning of the file."""
        return enum_missing(cls, value)

    def __str__(self):
        return str(self.value)

    @classmethod
    def from_directions(
        cls, directions: tuple[Direction, Direction]
    ) -> DiagonalDirection:
        """Get the diagonal direction from two cardinal directions. The two
        directions must be perpendicular."""

        if len(directions) != 2:
            raise ValueError("Exactly two directions are required.")
        if directions[0].to_orientation() == directions[1].to_orientation():
            raise ValueError(
                "The two directions must be perpendicular."
                f" Got {directions[0]} and {directions[1]}."
            )
        # Order should not matter
        directions_pair = frozenset(directions)
        for diag in cls:
            if frozenset(diag.components) == directions_pair:
                return diag
        raise ValueError(f"Invalid direction pair: {directions}.")

    @classmethod
    def from_vector(cls, vector: tuple[int, ...]) -> DiagonalDirection:
        """Get the diagonal direction from a 2D vector. The vector should have only
        two non-zero components in the first and the second dimension. The
        direction is determined by the sign of the components. The direction
        is returned as a DiagonalDirection enum.
        """
        if len(vector) != 2:
            raise ValueError("Vector must be 2D.")

        hor_coord, vert_coord = vector
        if hor_coord == 0 or vert_coord == 0:
            raise ValueError("Vector cannot have zero components.")

        hor_direction = Direction.from_vector((hor_coord, 0))
        vert_direction = Direction.from_vector((0, vert_coord))
        return DiagonalDirection.from_directions((hor_direction, vert_direction))

    def to_vector(self) -> tuple[int, int]:
        """Convert the diagonal direction to a 2D vector."""
        vectors = [direction.to_vector() for direction in self.components]
        # Sum the vectors to get the resulting vector
        return reduce(
            lambda acc, vec: (acc[0] + vec[0], acc[1] + vec[1]), vectors, (0, 0)
        )

    def opposite(self) -> DiagonalDirection:
        """Get the opposite diagonal direction."""
        opposite_directions = [direction.opposite() for direction in self.components]
        return DiagonalDirection.from_directions(opposite_directions)

    def direction_along_orientation(self, orientation: Orientation) -> Direction:
        """Get the cardinal direction along the given orientation."""
        return next(
            direction
            for direction in self.components
            if direction.to_orientation() == orientation
        )

    def mirror_across_orientation(self, orientation: Orientation) -> DiagonalDirection:
        """Get the mirrored diagonal direction across the given orientation."""
        return DiagonalDirection.from_directions(
            [
                direction.mirror_across_orientation(orientation)
                for direction in self.components
            ]
        )

    def __iter__(self) -> Iterable[Direction]:
        """Allow iteration over the component directions."""
        return iter(self.components)


class Orientation(str, Enum):
    """
    Orientation indicator for Operations (e.g. Split) and for Block initialization.
    """

    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"

    @classmethod
    def _missing_(cls, value):
        """
        Allow inputs with upper-case characters. For more details, see the
        documentation of `enum_missing` at the beginning of the file.
        """
        return enum_missing(cls, value)

    @classmethod
    def from_direction(cls, direction: Direction) -> Orientation:
        """
        Get the orientation from a direction.
        """
        match direction:
            case Direction.LEFT | Direction.RIGHT:
                return Orientation.HORIZONTAL
            case Direction.TOP | Direction.BOTTOM:
                return Orientation.VERTICAL
            case _:
                raise ValueError("Invalid direction. Cannot determine orientation.")

    @classmethod
    def from_vector(cls, vector: tuple[int, ...]) -> Orientation:
        """
        Get the orientation from a 2D vector. The orientation is determined
        by the direction of the vector. The orientation is returned as an
        Orientation enum.
        """
        return cls.from_direction(Direction.from_vector(vector))

    def perpendicular(self) -> Orientation:
        """
        Get the perpendicular orientation.
        """
        if self == Orientation.HORIZONTAL:
            return Orientation.VERTICAL
        if self == Orientation.VERTICAL:
            return Orientation.HORIZONTAL

        raise ValueError(f"Orientation has no perpendicular: {self}")


class ResourceState(str, Enum):
    """Supported states in state injection operations."""

    T = "t"
    S = "s"

    def get_expectation_value(self, basis: str) -> float:
        """Get the expectation value of the state in the Pauli basis of the state."""
        if basis not in ("X", "Y", "Z"):
            raise ValueError(
                f"Invalid basis: {basis}. Allowed values are 'X', 'Y', or 'Z'."
            )
        match self:
            case ResourceState.T:
                expectation = {
                    "Z": 0,  # 50% chance of measuring +1/-1 in Z basis
                    "X": cos(pi / 4),  # ~85% chance of measuring +1 in X basis
                    "Y": cos(pi / 4),  # ~85% chance of measuring +1 in Y basis
                }
            case ResourceState.S:
                expectation = {
                    "Z": 0,  # 50% chance of measuring +1/-1 in Z basis
                    "X": 0,  # 50% chance of measuring +1/-1 in X basis
                    "Y": 1,  # 100% chance of measuring +1 in Y basis
                }
            case _:
                raise ValueError(
                    f"Invalid resource state: {self}, allowed values are"
                    f"{list(ResourceState)}"
                )
        return float(expectation[basis])

    @classmethod
    def _missing_(cls, value):
        """Allow inputs with upper-case characters. For more details, see the
        documentation of `enum_missing` at the beginning of the file."""
        return enum_missing(cls, value)


class BoolOp(str, Enum):
    """
    BoolOp enum to specify operations on channels. Used in the "name" field of Circuit
    to indicate classical logic operations.

    It is understood that AND, NAND, OR, NOR, and XOR can correspond to operations that
    reduce multiple bits to a single bit, while MATCH and NOT correspond to an operation
    on one bit only.

    AND -> True when all channels are 1
    NAND -> True when at least one channel is 0
    OR -> True when at least one channel is 1
    NOR -> True when all channels are 0
    XOR -> True when an odd number of channels are 1

    NOT -> True when the channel is 0
    MATCH -> True when the channel is 1
    """

    # Multi-bit operation
    AND = "and"
    NAND = "nand"
    OR = "or"
    NOR = "nor"
    XOR = "xor"

    # Single-bit operation
    NOT = "not"
    MATCH = "match"

    @classmethod
    def _missing_(cls, value):
        """Allow inputs with upper-case characters. For more details, see the
        documentation of `enum_missing` at the beginning of the file."""
        return enum_missing(cls, value)

    def __str__(self):
        return str(self.value)

    @staticmethod
    def multi_bit_list() -> list[BoolOp]:
        """Return list of multi-bit operations."""
        return [BoolOp.AND, BoolOp.NAND, BoolOp.OR, BoolOp.NOR, BoolOp.XOR]

    @staticmethod
    def mono_bit_list() -> list[BoolOp]:
        """Return list of single-bit operations."""
        return [BoolOp.NOT, BoolOp.MATCH]



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from collections import defaultdict
from functools import partial, cached_property
import textwrap
from typing import TYPE_CHECKING, Any, Union

from pydantic import Field, field_validator


from ..eka import ChannelType, Circuit
from ..eka.utilities import BoolOp
from .op_signature import (
    ALL_EKA_OP_SIGNATURES,
    BOOL_LOGIC_OP_SIGNATURE,
    CLIFFORD_GATES_SIGNATURE,
    NONCLIFFORD_GATES_SIGNATURE,
    CONTROL_FLOW_OP_SIGNATURE,
    UTILS_SIGNATURE,
    OpType,
    OpSignature,
)
from .converter import Converter, OpAndTargetToInstrCallable

if TYPE_CHECKING:
    from hugr.qsystem.result import QsysResult, QsysShot

else:
    QsysResult = "QsysResult"  # pylint: disable=invalid-name
    QsysShot = "QsysShot"  # pylint: disable=invalid-name

ProgramAndRegister = tuple[str, dict[str, int], dict[str, int]]


class EkaToGuppylangConverter(Converter[ProgramAndRegister, Any]):
    """Convert Eka circuit format to Guppylang program.

    In the program, the quantum registers are represented as an array of qubits named
    `q`, and the classical registers are represented as an array of booleans named `c`.

    The program is generated as a string of Guppylang code, assuming the following
    import prefixes by default:

    - guppylang.std.quantum as `guppyQ.`
    - guppylang.std.qsystem as `guppyQSys.`
    - guppylang.std.builtins as `builtins.`

    Note: the user also needs to import `array` from `guppylang.std.builtins` to use
    arrays in the generated program (it is not exposed by the builtins by default).

    This can be customized via the constructor parameters.

    Parameters
    ----------
    quantum_prefix : str
        The import prefix for quantum operations in Guppylang.
    qsystem_prefix : str
        The import prefix for Qsys operations in Guppylang.
    builtins_prefix : str
        The import prefix for built-in functions in Guppylang.
    """

    SUPPORTED_OPERATIONS: frozenset[OpSignature] = ALL_EKA_OP_SIGNATURES

    ALLOW_ERROR_MODELS: bool = Field(default=False, frozen=True, init=False)

    quantum_prefix: str = Field(default="guppyQ.", frozen=True)
    qsystem_prefix: str = Field(default="guppyQSys.", frozen=True)
    builtins_prefix: str = Field(default="builtins.", frozen=True)

    @field_validator("quantum_prefix", "qsystem_prefix", "builtins_prefix")
    # pylint: disable=no-self-argument
    def validate_prefixes(cls, v):
        """Validate that the prefixes end with a dot."""
        cls._validate_import_prefix(v)
        return v

    @cached_property
    # pylint: disable=unused-argument
    def operations_map(
        self,
    ) -> dict[str, OpAndTargetToInstrCallable]:
        """Map operation names to their corresponding Guppy functions.
        The functions returned by this map take as input the quantum and classical
        targets, and return the Guppy function(s) to apply, handling assignments of
        parameters.

        Returns
        -------
        dict[str, str]
            A mapping from operation names to Guppylang instruction with the proper
            parameters applied.
        """

        def _single_qubit_op(
            q_targets: list[str],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            if op == "":
                return ""
            if isinstance(op, list):
                return "\n".join(f"{o}({q_targets[0]})" for o in op)
            return f"{op}({q_targets[0]})"

        def _two_qubit_op(
            q_targets: list[str],
            c_targets: list[str],
            op: str,
            desc: str = "",
        ) -> str:
            if op == "swap":
                return "\n".join(
                    [
                        f"{self.quantum_prefix}cx({q_targets[0]}, {q_targets[1]})",
                        f"{self.quantum_prefix}cx({q_targets[1]}, {q_targets[0]})",
                        f"{self.quantum_prefix}cx({q_targets[0]}, {q_targets[1]})",
                    ]
                )
            return f"{op}({', '.join(q_targets)})"

        def _measurement_op(
            q_targets: list[str],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            instructions = []
            if not isinstance(op, list):
                op = [op]
            for single_op in op[:-1]:
                instructions.append(f"{single_op}({q_targets[0]})")
            instructions.append(f"{c_targets[0]} = {op[-1]}({q_targets[0]})")
            instructions.append(
                f"{self.builtins_prefix}result('{desc}', {c_targets[0]})"
            )
            return "\n".join(instructions)

        def _control_flow_op(
            q_targets: list[str],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            """
            Generate Guppy code for classical control-flow constructs.
            str_instr is just a string like 'ifelse_circuit', 'if_circuit', etc.
            """
            match op:
                case "classical_if":
                    return f"if ({desc}):"
                case "classical_else":
                    return "else:"
                case "end_if":
                    return ""
                case _:
                    raise ValueError(f"Unsupported control flow operation: {op}")

        def _bool_logic_op(
            q_targets: tuple[str],
            c_targets: tuple[str],
            op: str,
            desc: str = "",
        ) -> str:
            condition = ""
            match op:
                case BoolOp.MATCH:
                    condition = c_targets[0]
                case BoolOp.NOT:
                    condition = f"not {c_targets[0]}"
                case BoolOp.AND:
                    condition = " and ".join(c_targets)
                case BoolOp.OR:
                    condition = " or ".join(c_targets)
                case BoolOp.XOR:
                    condition = " ^ ".join(c_targets)
                case BoolOp.NAND:
                    condition = f"not ({' and '.join(c_targets)})"
                case BoolOp.NOR:
                    condition = f"not ({' or '.join(c_targets)})"
                case _:
                    condition = ValueError(f"Unsupported BoolOp '{op}'.")
            return condition

        def _utils_op(
            q_targets: tuple[str],
            c_targets: tuple[str],
            op: str,
            desc: str = "",
        ) -> str:
            if op != "comment":
                raise ValueError(f"Unsupported utils operation: {op}")
            return f"# {desc}"

        # Map operation types to their corresponding application function
        op_type_handlers = {
            OpType.SINGLE_QUBIT: _single_qubit_op,
            OpType.TWO_QUBIT: _two_qubit_op,
            OpType.MEASUREMENT: _measurement_op,
            OpType.RESET: _single_qubit_op,
            OpType.CONTROL_FLOW: _control_flow_op,
            OpType.BOOL_LOGIC: _bool_logic_op,
            OpType.UTILS: _utils_op,
        }

        eka_to_guppy_ops = {
            "i": "",
            "x": f"{self.quantum_prefix}x",
            "y": f"{self.quantum_prefix}y",
            "z": f"{self.quantum_prefix}z",
            "h": f"{self.quantum_prefix}h",
            "t": f"{self.quantum_prefix}t",
            "phase": f"{self.quantum_prefix}s",
            "phaseinv": f"{self.quantum_prefix}sdg",
            "cnot": f"{self.quantum_prefix}cx",
            "cy": f"{self.quantum_prefix}cy",
            "cz": f"{self.quantum_prefix}cz",
            "cx": f"{self.quantum_prefix}cx",
            "swap": "swap",
            "reset": f"{self.quantum_prefix}reset",
            "reset_0": f"{self.quantum_prefix}reset",
            "reset_1": [f"{self.quantum_prefix}reset", f"{self.quantum_prefix}x"],
            "reset_+": [f"{self.quantum_prefix}reset", f"{self.quantum_prefix}h"],
            "reset_-": [
                f"{self.quantum_prefix}reset",
                f"{self.quantum_prefix}x",
                f"{self.quantum_prefix}h",
            ],
            "reset_+i": [
                f"{self.quantum_prefix}reset",
                f"{self.quantum_prefix}h",
                f"{self.quantum_prefix}s",
            ],
            "reset_-i": [
                f"{self.quantum_prefix}reset",
                f"{self.quantum_prefix}x",
                f"{self.quantum_prefix}h",
                f"{self.quantum_prefix}s",
            ],
            "measurement": f"{self.qsystem_prefix}measure_and_reset",
            "measure_z": f"{self.qsystem_prefix}measure_and_reset",
            "measure_x": [
                f"{self.quantum_prefix}h",
                f"{self.qsystem_prefix}measure_and_reset",
            ],
            "measure_y": [
                f"{self.quantum_prefix}sdg",
                f"{self.quantum_prefix}h",
                f"{self.qsystem_prefix}measure_and_reset",
            ],
        }

        return {
            op.name: partial(
                op_type_handlers[op.op_type],
                op=eka_to_guppy_ops[op.name],
            )
            for op in CLIFFORD_GATES_SIGNATURE | NONCLIFFORD_GATES_SIGNATURE
        } | {
            op.name: partial(op_type_handlers[op.op_type], op=op.name)
            for op in CONTROL_FLOW_OP_SIGNATURE
            | UTILS_SIGNATURE
            | BOOL_LOGIC_OP_SIGNATURE
        }

    def convert_circuit(self, input_circuit: Circuit) -> str:
        """Convert a Circuit to Guppy program string.

        Parameters
        ----------
        input_circuit : Circuit
            The Eka Circuit to convert.

        Returns
        -------
        str
            The Guppy program string representing the converted circuit.
        """

        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit")

        # Return an empty program if there are no channels or the circuit is empty.
        if (
            not input_circuit.circuit and input_circuit.name not in self.operations_map
        ) or not input_circuit.channels:
            return "", {}, {}

        init_program_line, q_registers, c_registers = self.emit_init_instructions(
            input_circuit
        )

        program_lines = []
        program_lines.extend(init_program_line.splitlines())

        circuit_instructions = self.emit_circuit_program(
            input_circuit, q_registers, c_registers
        )
        program_lines.extend(circuit_instructions.splitlines())

        # Free the quantum resources at the end of the program
        # (this is required in Guppylang)
        program_lines.append("for e in q:")
        program_lines.append(f"   {self.quantum_prefix}discard(e) ")

        return (
            "\n".join(program_lines),
            q_registers,
            c_registers,
        )

    @staticmethod
    def parse_target_run_outcome(
        outcome: Union[QsysResult, QsysShot],
    ) -> dict[str, list[int]]:
        """
        Convert a QsysResult into a dict mapping label -> list of int per shot.
        """
        result_dict = defaultdict(list)

        # Use duck typing to avoid runtime dependency on hugr types
        if hasattr(outcome, "results"):
            # QsysResult case: has 'results' attribute containing shots
            for shot in outcome.results:
                for label, value in shot.entries:
                    result_dict[label].append(value)
        elif hasattr(outcome, "entries"):
            # QsysShot case: has 'entries' attribute directly
            for label, value in outcome.entries:
                result_dict[label].append(value)
        return dict(result_dict)

    def emit_init_instructions(
        self,
        circuit: Circuit,
    ) -> tuple[str, dict[str, str], dict[str, str]]:
        """Generate Guppy code to initialize quantum and classical registers.

        Parameters
        ----------
        circuit : Circuit
            The circuit for which to initialize registers.

        Returns
        -------
        tuple[str, dict[str, str], dict[str, str]]
            A tuple containing:

            - The Guppy code string to initialize the registers.
            - A mapping from quantum channel IDs to a string of the variable name in
              the program (e.g. "q[0]").
            - A mapping from classical channel IDs to a string of the variable name in
              the program (e.g. "c[0]").

        """

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        q_channels = sorted(
            [chan for chan in circuit.channels if chan.type == ChannelType.QUANTUM],
            key=lambda ch: ch.label,
        )
        qubit_id_to_idx_map = {v.id: f"q[{i}]" for i, v in enumerate(q_channels)}

        c_channels = sorted(
            [chan for chan in circuit.channels if chan.type == ChannelType.CLASSICAL],
            key=lambda ch: ch.label,
        )
        creg_id_to_idx_map = {v.id: f"c[{i}]" for i, v in enumerate(c_channels)}

        q_len = len(q_channels)
        c_len = len(c_channels)

        # pylint: disable=line-too-long
        init_code = "\n".join(
            [
                f"q = array({self.quantum_prefix}qubit() for _ in range({q_len}))",
                f"c = array(False for _ in range({c_len}))",
            ]
        )

        # Remove common leading whitespace and top/bottom empty lines
        init_code_clean = textwrap.dedent(init_code).strip()

        return init_code_clean, qubit_id_to_idx_map, creg_id_to_idx_map

    def emit_leaf_circuit_instruction(
        self,
        circuit: Circuit,
        quantum_channel_map: dict[str, str],
        classical_channel_map: dict[str, str],
    ) -> str:
        """Generate a string of the Guppy code of an given Eka instruction.

        Parameters
        ----------
        circuit : Circuit
            The Eka atomic circuit to convert to Guppy code.
        quantum_channel_map : dict[str, str]
            A mapping from quantum channel IDs to their variable names in the Guppylang
            code.
        classical_channel_map : dict[str, str]
            A mapping from classical channel IDs to their variable names in the
            Guppylang code.

        Returns
        -------
        str
            The Guppy code representation of the Eka instruction.

        Raises
        ------
        ValueError
            If the circuit is not atomic (leaf in the Eka circuit tree).
        KeyError
            If the instruction is not supported by the converter.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if circuit.circuit and len(circuit.circuit) != 0:
            raise ValueError("The circuit must be an leaf circuit with channels.")

        if circuit.name not in self.operations_map:
            raise ValueError(f"Unsupported operation: {circuit.name}")

        # Get quantum targets
        q_targets = [
            quantum_channel_map[q_chan.id]
            for q_chan in circuit.channels
            if q_chan.type == ChannelType.QUANTUM
        ]

        # Get classical targets
        c_targets = [
            classical_channel_map[c_chan.id]
            for c_chan in circuit.channels
            if c_chan.type == ChannelType.CLASSICAL
        ]

        self._validate_ops_args(circuit.name, len(q_targets), len(c_targets))

        c_label = [
            c_chan.label
            for c_chan in circuit.channels
            if c_chan.type == ChannelType.CLASSICAL
        ]

        if self.op_by_eka_name[circuit.name].op_type == OpType.MEASUREMENT:
            desc = c_label[0] if c_label else ""
        else:
            desc = circuit.description

        return self.operations_map[circuit.name](q_targets, c_targets, desc=desc)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import json

from pydantic.dataclasses import dataclass, Field
from loom.eka.utilities import dataclass_config

from .syndrome import Syndrome
from .utilities import Cbit


@dataclass(config=dataclass_config)
class Detector:
    """
    A detector is the parity of multiple syndromes. This dataclass does not store the
    actual value of a detector but instead the required information on how to calculate
    it given the output of a quantum circuit.

    Parameters
    ----------
    syndromes : tuple[Syndrome, ...]
        Syndromes that this detector corresponds to, in temporal order of measurement.
        NOTE: "Please ensure this order for proper functioning of the detector"
    """

    syndromes: tuple[Syndrome, ...]
    labels: dict[str, str | tuple[int, ...] | int] = Field(default_factory=dict)

    @property
    def measurements(self) -> tuple[Cbit, ...]:
        """
        Returns the measurements of the constituent syndromes. The corrections
        must be used in a detector definition only once.

        Returns
        -------
        tuple[Cbit, ...]
            The detector measurements
        """
        measurements_list = []
        for syndrome in self.syndromes:
            measurements_list.extend(syndrome.measurements)

        # only include corrections from the last syndrome
        measurements_list.extend(self.syndromes[-1].corrections)
        return tuple(measurements_list)

    def dumps(self) -> str:
        """
        Return the detector as a JSON string. This string does not contain the
        complete Syndrome objects but only their UUIDs to not store redundant
        information.
        """
        self_dict = {"syndromes": [syndrome.uuid for syndrome in self.syndromes]}
        return json.dumps(self_dict)

    def rounds(self) -> tuple[int]:
        """
        Returns the rounds of the constituent syndromes

        Returns
        -------
        tuple[int]
            The detector rounds
        """
        rounds_list = []
        for syndrome in self.syndromes:
            rounds_list.append(syndrome.round)

        return tuple(set(rounds_list))

    def stabilizer(self) -> tuple[str]:
        """
        Returns the UUIDs of the stabilizer(s) of the constituent syndromes

        Returns
        -------
        tuple[str]
            The UUIDs of the stabilizer(s) of the constituent syndromes.
        """
        stab_ids = tuple(syndrome.stabilizer for syndrome in self.syndromes)
        return stab_ids

    def __eq__(self, other: "Detector") -> bool:
        if not isinstance(other, Detector):
            raise NotImplementedError(
                f"Cannot compare Detector with {type(other)} object."
            )
        return set(self.syndromes) == set(other.syndromes)

    def __repr__(self) -> str:
        syndrome_repr = ", ".join(repr(syndrome) for syndrome in self.syndromes)
        return f"Detector(Syndromes: ({syndrome_repr}), Labels: {self.labels})"

    def __hash__(self):
        return hash(self.syndromes)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import bisect
import uuid
from collections.abc import Iterator

from pydantic.dataclasses import dataclass


class BlocksAlreadySeenError(Exception):
    """
    Exception raised when attempting to add blocks that have already been seen
    in the BlockHistory.

    Attributes
    ----------
    blocks : frozenset[str]
        The set of block UUIDs that have already been seen.
    """

    def __init__(self, blocks: set[str]) -> None:
        super().__init__(f"Blocks already seen in BlockHistory: {blocks}")
        self.blocks = frozenset(blocks)


class BlocksNotPresentError(Exception):
    """
    Exception raised when attempting to remove blocks that are not present
    in the BlockHistory at a given timestamp.

    Attributes
    ----------
    blocks : frozenset[str]
        The set of block UUIDs that are not present.
    timestamp : int
        The timestamp at which the blocks were expected to be present.
    """

    def __init__(self, blocks: set[str], timestamp: int) -> None:
        super().__init__(f"Blocks not present at timestamp {timestamp}: {blocks}")
        self.blocks = frozenset(blocks)
        self.timestamp = timestamp


class InconsistentBlockUpdateError(Exception):
    """
    Exception raised when an inconsistent block update is detected while
    propagating changes forward in time. This indicates that the expected blocks
    were not present at a future timestamp.

    Attributes
    ----------
    blocks : frozenset[str]
        The set of block UUIDs that caused the inconsistency.
    timestamp : int
        The timestamp at which the inconsistency was detected.
    """

    def __init__(self, blocks: set[str], timestamp: int) -> None:
        super().__init__(
            f"Inconsistent block update detected at timestamp {timestamp}. "
            f"Missing blocks: {blocks}"
        )
        self.blocks = frozenset(blocks)
        self.timestamp = timestamp


@dataclass
class BlockHistory:
    """
    Class to track the history of blocks over time.

    Attributes
    ----------
    _blocks_by_timestamp : dict[int, set[str]]
        A mapping from timestamps to sets of block UUIDs present at those times.
    _timestamps_sorted_asc : list[int]
        A sorted list of timestamps in ascending order. This allows for efficient
        searching of previous and next timestamps.
    _timestamps_set : set[int]
        A set of timestamps for quick existence checks.
    _all_blocks_set : set[str]
        A set of all block UUIDs that have ever been present in the history.
    """

    _blocks_by_timestamp: dict[int, set[str]]
    _timestamps_sorted_asc: list[int]
    _timestamps_set: set[int]
    _all_blocks_set: set[str]

    @classmethod
    def create(cls, blocks_at_0: set[str]) -> "BlockHistory":
        """
        Create a BlockHistory instance with initial blocks at timestamp 0.

        Parameters
        ----------
        blocks_at_0 : set[str]
            Set of block UUIDs present at timestamp 0.

        Returns
        -------
        BlockHistory
            An instance of BlockHistory initialized with the given blocks at time 0.
        """
        if not cls.is_set_of_uuid4(blocks_at_0):
            raise ValueError("blocks_at_0 must be a set of valid UUID4 strings.")
        return cls(
            _blocks_by_timestamp={0: blocks_at_0.copy()},
            _timestamps_sorted_asc=[0],
            _timestamps_set={0},
            _all_blocks_set=blocks_at_0.copy(),
        )

    def max_timestamp_below_ref_value(self, ref_timestamp: int) -> int:
        """
        Return the largest timestamp less than the input timestamp that exists
        in the block history.

        Parameters
        ----------
        ref_timestamp : int
            The reference timestamp.

        Returns
        -------
        int
            The largest timestamp less than the input timestamp.
        """
        self.validate_timestamp(ref_timestamp)
        if ref_timestamp == 0:
            raise IndexError("No previous timestamp exists for timestamp 0.")

        idx = bisect.bisect_left(self._timestamps_sorted_asc, ref_timestamp)
        return self._timestamps_sorted_asc[idx - 1]

    def min_timestamp_above_ref_value(self, ref_timestamp: int) -> int | None:
        """
        Return the smallest timestamp greater than the input timestamp that exists
        in the block history.

        Parameters
        ----------
        ref_timestamp : int
            The reference timestamp.

        Returns
        -------
        int | None
            The smallest timestamp greater than the input timestamp, or None if
            no such timestamp exists.
        """
        self.validate_timestamp(ref_timestamp)

        # Use bisect to find the right position and retrieve the next timestamp
        idx = bisect.bisect_right(self._timestamps_sorted_asc, ref_timestamp)
        return (
            self._timestamps_sorted_asc[idx]
            if idx < len(self._timestamps_sorted_asc)
            else None
        )

    def block_uuids_at_index(self, index: int) -> set[str]:
        """
        Return the blocks present at the timestamp corresponding to the given index
        in the sorted list of timestamps.

        Parameters
        ----------
        index : int
            The index in the sorted list of timestamps. Can be negative for reverse
            indexing.

        Returns
        -------
        set[str]
            The set of block UUIDs present at the timestamp corresponding to the index.
        """
        if not isinstance(index, int):
            raise ValueError(f"Index must be an integer. Got {type(index).__name__}.")
        if not (
            -len(self._timestamps_sorted_asc)
            <= index
            < len(self._timestamps_sorted_asc)
        ):
            raise IndexError("Index out of range for timestamps.")

        timestamp = self._timestamps_sorted_asc[index]
        return self._blocks_by_timestamp[timestamp].copy()

    def blocks_at(self, timestamp: int) -> set[str]:
        """
        Return the blocks present at the given timestamp. If the exact timestamp
        does not exist, return the blocks at the most recent previous timestamp.

        Parameters
        ----------
        timestamp : int
            The timestamp to query.

        Returns
        -------
        set[str]
            The set of block UUIDs present at the given timestamp.
        """
        self.validate_timestamp(timestamp)

        timestamp = (
            timestamp
            if timestamp in self._timestamps_set
            else self.max_timestamp_below_ref_value(timestamp)
        )
        return self._blocks_by_timestamp[timestamp].copy()

    def blocks_over_time(
        self, t_start: int | None = None, t_stop: int | None = None
    ) -> Iterator[tuple[int, set[str]]]:
        """
        Return the timestamps and corresponding blocks over the specified time range.

        Parameters
        ----------
        t_start : int | None
            The start timestamp (inclusive). If None, starts from the first timestamp.
        t_stop : int | None
            The stop timestamp (exclusive). If None, goes up to the last timestamp.

        Returns
        -------
        Iterator[tuple[int, set[str]]]
            An iterator over (timestamp, set of block UUIDs) tuples within the specified
            time range.
        """
        # Determine start and stop timestamps
        t_start = t_start if t_start is not None else self._timestamps_sorted_asc[0]
        t_stop = t_stop if t_stop is not None else self._timestamps_sorted_asc[-1] + 1

        self.validate_timestamp(t_start)
        self.validate_timestamp(t_stop)

        # Find indices for slicing using bisect
        start_idx = bisect.bisect_left(self._timestamps_sorted_asc, t_start)
        stop_idx = bisect.bisect_left(self._timestamps_sorted_asc, t_stop)

        for idx in range(start_idx, stop_idx):
            timestamp = self._timestamps_sorted_asc[idx]
            # Return a copy of the block set to prevent external modification,
            # since sets are mutable
            yield timestamp, self._blocks_by_timestamp[timestamp].copy()

    def update_blocks_MUT(  # pylint: disable=invalid-name
        self, timestamp: int, old_blocks: set[str], new_blocks: set[str]
    ) -> None:
        """
        Modify the blocks at a given timestamp by replacing old_blocks with new_blocks.
        If the timestamp does not exist, it is created based on the previous timestamp's
        blocks.

        Parameters
        ----------
        timestamp : int
            The timestamp at which to modify the blocks.
        old_blocks : set[str]
            The set of block UUIDs to be removed.
        new_blocks : set[str]
            The set of block UUIDs to be added.
        """
        # Validate timestamp
        self.validate_timestamp(timestamp)
        # Check that all old_blocks and new_blocks are sets of UUIDs (strings)
        if not self.is_set_of_uuid4(old_blocks):
            raise ValueError("old_blocks must be a set of valid UUID4 strings.")
        if not self.is_set_of_uuid4(new_blocks):
            raise ValueError("new_blocks must be a set of valid UUID4 strings.")

        # Ensure that new_blocks have not been seen before
        blocks_seen_before = new_blocks.intersection(self._all_blocks_set)
        if blocks_seen_before:
            raise BlocksAlreadySeenError(blocks_seen_before)

        # Update the set of all blocks ever seen
        self._all_blocks_set.update(new_blocks)

        # Check whether timestamp exists
        timestamp_exists = timestamp in self._timestamps_set

        # Determine which block set to modify
        blocks_to_modify = self._blocks_by_timestamp[
            (
                timestamp
                if timestamp_exists
                else self.max_timestamp_below_ref_value(timestamp)
            )
        ]

        # Ensure that old_blocks are present in blocks_to_modify
        if not old_blocks.issubset(blocks_to_modify):
            missing_blocks = old_blocks - blocks_to_modify
            raise BlocksNotPresentError(missing_blocks, timestamp)

        # Compute updated blocks for this timestamp
        updated_blocks = (blocks_to_modify - old_blocks) | new_blocks

        # Update the blocks at the relevant timestamp
        self._blocks_by_timestamp[timestamp] = updated_blocks

        if not timestamp_exists:
            # If timestamp did not exist, insert it into sorted list and set
            bisect.insort(self._timestamps_sorted_asc, timestamp)
            self._timestamps_set.add(timestamp)

        # For operations that run in parallel, there is a chance we might be modifying
        # blocks at a timestamp that is not the latest one.
        # Since timestamp lies in the past, we need to propagate the changes forward
        # to ensure consistency.
        # NOTE: The operations running in parallel with the current one cannot have
        # modified `old_blocks` at any future timestamp since parallel operations
        # cannot have overlapping Block objects.
        for t, block_set in self.blocks_over_time(t_start=timestamp + 1):
            if not old_blocks.issubset(block_set):
                missing_blocks = old_blocks - block_set
                raise InconsistentBlockUpdateError(missing_blocks, t)
            # Update all subsequent timestamps to reflect the changes
            block_difference = block_set - old_blocks
            self._blocks_by_timestamp[t] = block_difference | new_blocks

    # Utilities for Validation of Inputs

    @staticmethod
    def validate_timestamp(timestamp: int) -> None:
        """
        Raise ValueError if timestamp is not a non-negative integer.
        """
        if not isinstance(timestamp, int) or timestamp < 0:
            raise ValueError(
                f"Timestamp must be a non-negative integer. Got {timestamp} of type "
                f"{type(timestamp).__name__} instead."
            )

    @staticmethod
    def is_uuid4(s: str) -> bool:
        """
        Return True if s is a valid UUID4 string, False otherwise.
        """
        try:
            u = uuid.UUID(s)
        except ValueError:
            return False
        return u.version == 4

    @classmethod
    def is_set_of_uuid4(cls, blocks: set[str]) -> bool:
        """
        Return True if blocks is a set of valid UUID4 strings, False otherwise.
        """
        return isinstance(blocks, set) and all(cls.is_uuid4(b) for b in blocks)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from enum import Enum
from typing import Optional
from uuid import uuid4

from pydantic import Field, field_validator, ValidationInfo
from pydantic.dataclasses import dataclass

from .utilities.validation_tools import (
    dataclass_config,
    uuid_error,
    ValidationInfo,
)


class ChannelType(str, Enum):
    """
    The type of the channel: QUANTUM or CLASSICAL
    More types should be added when we feel the need for it
    """

    QUANTUM = "quantum"
    CLASSICAL = "classical"


def create_default_label(channel_type: ChannelType):
    """
    Creates a default label for the channel.

    Parameters
    ----------
    channel_type : ChannelType
        The type of the channel: QUANTUM or CLASSICAL

    Returns
    -------
    str
        The default label for the channel
    """

    match channel_type:
        case ChannelType.QUANTUM:
            return "data_qubit"
        case ChannelType.CLASSICAL:
            return "classical_bit"
        case _:
            raise ValueError(f"Channel type {type} not recognized")


@dataclass(config=dataclass_config)
class Channel:
    """
    Identifies information channels connecting the Circuit elements: examples are
    classical or quantum bit channels

    Parameter
    ---------
    type: ChannelType
        The type of the channel: QUANTUM or CLASSICAL, default is QUANTUM

    label: str
        The label of the channel, allowing it to be grouped in a user friendly way, E.g.
        can be "red", "ancilla_qubit" or "my_favourite_qubit"

    id: str
        The unique identifier of the channel
    """

    type: ChannelType = Field(default=ChannelType.QUANTUM)
    label: Optional[str] = Field(default=None, validate_default=True)
    id: str = Field(default_factory=lambda: str(uuid4()))

    # Validation functions
    _validate_uuid = field_validator("id")(uuid_error)

    @field_validator("label", mode="after")
    @classmethod
    def set_default_label(cls, v: str, info: ValidationInfo) -> str:
        """
        Set the default label based on the type of the channel, according to
        the following scheme:
        ChannelType.QUANTUM:   "data_qubit"
        ChannelType.CLASSICAL: "classical_bit"
        """
        if v is None and "type" in info.data.keys():
            v = create_default_label(info.data["type"])
        return v

    # Magic methods
    def __eq__(self, other):
        if isinstance(other, Channel):
            return (self.type, self.id) == (other.type, other.id)
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.type, self.id))

    # Convenience methods

    def is_quantum(self) -> bool:
        """Check if the channel is a quantum channel.

        Returns
        -------
        bool
            True if the channel is a quantum channel, False otherwise.
        """
        return self.type == ChannelType.QUANTUM

    def is_classical(self) -> bool:
        """Check if the channel is a classical channel.
        Returns
        -------
        bool
            True if the channel is a classical channel, False otherwise.
        """
        return self.type == ChannelType.CLASSICAL



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Eka
from loom.interpreter.applicator import CodeApplicator

from ..code_factory import ShorCode


class ShorCodeApplicator(CodeApplicator):  # pylint: disable=too-few-public-methods
    """
    Contains the implementation logic for each operation, for the Shor code.
    """

    def __init__(
        self,
        eka: Eka,
    ):
        # Ensure that all blocks are typed ShorCode
        if any(not isinstance(block, ShorCode) for block in eka.blocks):
            raise ValueError("All blocks must be of type ShorCode.")
        super().__init__(eka)
        # Add the extra operations that are supported by the Shor Code
        self.supported_operations |= {}



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from abc import ABC

import numpy as np


class PauliArray(ABC):
    """
    Abstract class for PauliArray, parent of StabArray and Tableau.

    Parameters
    ----------
    array : np.ndarray
        The array representation of the PauliArray.
    """

    array: np.ndarray

    @property
    def nqubits(self) -> int:
        """
        The number of qubits that the PauliArray operators act on.
        """
        return self.array.shape[1] // 2

    @property
    def x(self) -> np.ndarray:
        """
        The array representing the X-component of the PauliArray in binary
        representation.
        """
        return self.array[:, : self.nqubits]

    @property
    def z(self) -> np.ndarray:
        """
        The array representing the Z-component of the PauliArray in binary
        representation.
        """
        return self.array[:, self.nqubits : 2 * self.nqubits]



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from itertools import product

from loom.eka import Block, Circuit, PauliOperator, Stabilizer
from loom.eka.utilities import Orientation
from loom.eka.operations import (
    Merge,
    MeasureBlockSyndromes,
    LogicalMeasurement,
)
from loom.interpreter import InterpretationStep, Cbit, Syndrome
from loom.interpreter.applicator import generate_syndromes, measureblocksyndromes

from ..code_factory import RotatedSurfaceCode


# pylint: disable=too-many-lines
def merge_consistency_check(  # pylint: disable=too-many-branches
    interpretation_step: InterpretationStep,
    operation: Merge,
) -> tuple[RotatedSurfaceCode, RotatedSurfaceCode]:
    """
    Check if the blocks are compatible for merging. Also re-orders the blocks so
    that block1 is the left or top block and block2 is the right or bottom block. This
    is required for the rest of the workflow to function properly.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        InterpretationStep containing the blocks to merge.
    operation : Merge
        Descriptor of the merge operation to perform.

    Returns
    -------
    tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        The blocks to merge, ordered so that block1 is the left or top block and
        block2 is the right or bottom block.

    Raises
    ------
    ValueError
        If the blocks are not compatible for merging. The blocks are compatible if:
        (1) They do not overlap (this should already be enforced).
        (2) The blocks' upper left corner qubits are aligned (either horizontally or
        vertically).
        (3) The blocks have the same size in the direction normal to the merge.
        (4) The boundaries to be merged are of the same type.
        (5) The alternate pattern of stabilizers is preserved. I.e. the blocks have the
        same upper left 4-body stabilizer if the distance is even, different stabilizer
        if the distance is odd.
        (6) There needs to be at least one row/column of data qubits between the two
        blocks.
        (7) The blocks have a single logical operator.
        Also raises a ValueError if the orientation specified in the operation does not
        match the default orientation.
    """

    name1, name2 = operation.input_blocks_name
    block1 = interpretation_step.get_block(name1)
    block2 = interpretation_step.get_block(name2)
    # 1 - Check that the blocks are of the right type
    if not isinstance(block1, RotatedSurfaceCode) or not isinstance(
        block2, RotatedSurfaceCode
    ):
        raise TypeError(
            f"The merge operation is not supported for "
            f"{tuple(set(type(block1), type(block2)))} blocks."
        )
    context_str = (
        # pylint: disable=protected-access
        f"Operation {operation.__class__.__name__} on {operation._inputs} failed:\n"
    )

    # 2 - Check that the blocks do not overlap
    if len(set(block1.data_qubits + block2.data_qubits)) < len(
        block1.data_qubits
    ) + len(block2.data_qubits):
        raise ValueError(context_str + "The blocks overlap.")

    # 3 - Check that the blocks' upper left corners are aligned
    # Compute the number of non zero coordinates in the vector that links the upper left
    # corners of the two blocks.
    link_vector = [
        coord2 - coord1
        for coord1, coord2 in zip(
            block1.upper_left_qubit, block2.upper_left_qubit, strict=True
        )
    ]
    blocks_are_aligned = sum(coord != 0 for coord in link_vector) == 1
    index_nonzero = next((i for i, coord in enumerate(link_vector) if coord != 0), None)

    if not blocks_are_aligned:
        raise ValueError(
            context_str + "The blocks' upper left corners are not aligned."
        )

    # Ensure that block1 is left or top and block2 is right or bottom
    if link_vector[index_nonzero] < 0:
        block1, block2 = block2, block1

    # Find the direction of the merge
    if index_nonzero == 0:
        default_orientation = Orientation.HORIZONTAL
        merge_is_horizontal = True
    elif index_nonzero == 1:
        default_orientation = Orientation.VERTICAL
        merge_is_horizontal = False
    else:
        raise ValueError(
            context_str
            + "The blocks corners have same x and y coordinates but differ in another "
            "way."
        )

    orientation = operation.orientation or default_orientation
    if orientation != default_orientation:
        raise ValueError(
            context_str + "The orientation specified in the operation does not "
            "match the default orientation."
        )

    # 4 - Check that the blocks have the same size in the direction normal to the merge
    if merge_is_horizontal:
        if block1.size[1] != block2.size[1]:
            raise ValueError(
                context_str
                + "The blocks have different sizes in the vertical direction."
            )
    else:
        if block1.size[0] != block2.size[0]:
            raise ValueError(
                context_str
                + "The blocks have different sizes in the horizontal direction."
            )

    # 5 - Check that the boundaries to be merged are of the same type
    if merge_is_horizontal:
        if block1.boundary_type("right") != block2.boundary_type("left"):
            raise ValueError(
                context_str + "The boundaries to be merged are of different types."
            )
    else:
        if block1.boundary_type("bottom") != block2.boundary_type("top"):
            raise ValueError(
                context_str + "The boundaries to be merged are of different types."
            )

    # 6 - Check that the alternate pattern of stabilizers is preserved
    if merge_is_horizontal:
        distance = block2.upper_left_qubit[0] - block1.upper_left_qubit[0]
    else:
        distance = block2.upper_left_qubit[1] - block1.upper_left_qubit[1]
    #  If distance is even, upper left 4-body stabilizers should be the same type
    #  If distance is odd, upper left 4-body stabilizers should be different types
    if (
        distance % 2 == 0
        and (
            block1.upper_left_4body_stabilizer.pauli
            != block2.upper_left_4body_stabilizer.pauli
        )
    ) or (
        distance % 2 != 0
        and (
            block1.upper_left_4body_stabilizer.pauli
            == block2.upper_left_4body_stabilizer.pauli
        )
    ):
        raise ValueError(
            context_str + "The alternate pattern of stabilizers is not preserved."
        )

    # 7 - Check that there is at least one row/column of data qubits between the two
    # blocks
    if merge_is_horizontal:
        if block1.upper_left_qubit[0] + block1.size[0] == block2.upper_left_qubit[0]:
            raise ValueError(
                context_str
                + "There is no column of data qubits between the two blocks."
            )
    else:
        if block1.upper_left_qubit[1] + block1.size[1] == block2.upper_left_qubit[1]:
            raise ValueError(
                context_str + "There is no row of data qubits between the two blocks."
            )

    return (block1, block2)


def create_merge_circuit(
    interpretation_step: InterpretationStep,
    blocks: tuple[RotatedSurfaceCode, RotatedSurfaceCode],
    operation: Merge,
    qubits_to_reset: tuple[tuple[int, ...], ...],
    boundary_type: str,
) -> Circuit:
    """Create the circuit for the merge operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks to merge.
    blocks : tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        Blocks to merge.
    operation : Merge
        Descriptor of the merge operation to perform.
    qubits_to_reset : tuple[tuple[int, ...], ...]
        Qubits to reset during the merge operation.
    boundary_type : str
        Type of boundary to merge.

    Returns
    -------
    Circuit
        Circuit for the merge operation.
    """
    # B) - CIRCUIT
    #    B.1) Find which basis the data qubits have to be reset in
    # If the boundary type is X the stabilizers are Z type, they need to be reset in |0>
    # If the boundary type is Z the stabilizers are X type, they need to be reset in |+>
    reset_state = "0" if boundary_type == "X" else "+"
    #    B.2) Create a reset circuit for every reset data qubit
    reset_circuit_seq = [
        [
            Circuit(
                f"Reset_{reset_state}",
                channels=[interpretation_step.get_channel_MUT(q)],
            )
            for q in qubits_to_reset
        ]
    ]

    merge_circuit = Circuit(
        name=(
            f"Merge {blocks[0].unique_label} and {blocks[1].unique_label} into "
            f"{operation.output_block_name}"
        ),
        circuit=reset_circuit_seq,
    )

    return merge_circuit


def find_merge_stabilizer_to_circuit_mappings(
    blocks: tuple[RotatedSurfaceCode, RotatedSurfaceCode],
    new_stabs_left: list[Stabilizer],
    new_stabs_right: list[Stabilizer],
    new_stabs_top: list[Stabilizer],
    new_stabs_bottom: list[Stabilizer],
    new_bulk_stabilizers: list[Stabilizer],
    old_stabs_to_lengthen: list[Stabilizer],
) -> dict[str, str]:
    """Finds the mapping between the stabilizers in the new block and the associated
    syndrome circuits.

    Parameters
    ----------
    blocks : tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        Blocks to merge.
    new_stabs_left : list[Stabilizer]
        New stabilizers that form a part of the left boundary of the new block. This is
        an empty list for horizontal merges.
    new_stabs_right : list[Stabilizer]
        New stabilizers that form a part of the right boundary of the new block. This is
        an empty list for horizontal merges.
    new_stabs_top : list[Stabilizer]
        New stabilizers that form a part of the top boundary of the new block. This is
        an empty list for vertical merges.
    new_stabs_bottom : list[Stabilizer]
        New stabilizers that form a part of the bottom boundary of the new block. This
        is an empty list for vertical merges.
    new_bulk_stabilizers : list[Stabilizer]
        New stabilizers that form a part of the bulk of the new block.
    old_stabs_to_lengthen : list[Stabilizer]
        Old stabilizers that are gonna be replaced by new stabilizers.

    Returns
    -------
    dict[str, str]
        New mapping of stabilizers to syndrome circuits for the block.
    """
    block1, block2 = blocks
    # We only use one set of SyndromeCircuits for the new block
    # Create the mapping from block2 syndrome circuits to the new syndrome_circuits list
    block2_to_block1_synd_circ_map = {
        block2_synd_circ.uuid: block1_synd_circ.uuid
        for block1_synd_circ in block1.syndrome_circuits
        for block2_synd_circ in block2.syndrome_circuits
        if block1_synd_circ.name == block2_synd_circ.name
    }
    # Ensure that the id corresponds to a stabilizer that is not removed
    conserved_stabs_id = [
        stab.uuid
        for stab in block1.stabilizers + block2.stabilizers
        if stab not in old_stabs_to_lengthen
    ]
    # Use the stabilizer to circuit mapping of the first block (apart from removed
    # stabilizers)
    new_stab_to_circuit = {
        stab_id: synd_circ_id
        for (stab_id, synd_circ_id) in block1.stabilizer_to_circuit.items()
        if stab_id in conserved_stabs_id
    }
    # Create a new mapping to the syndrome circuits for the new block for the
    # stabilizers of block2 that are not gonna be replaced
    new_stab_to_circuit.update(
        {
            stab_id: block2_to_block1_synd_circ_map[synd_circ_id]
            for (stab_id, synd_circ_id) in block2.stabilizer_to_circuit.items()
            if stab_id in conserved_stabs_id
        }
    )

    # Left stabilizers
    for stab in new_stabs_left:
        new_stab_to_circuit[stab.uuid] = next(
            syndrome_circuit.uuid
            for syndrome_circuit in block1.syndrome_circuits + block2.syndrome_circuits
            if syndrome_circuit.name == f"left-{stab.pauli.lower()}"
        )
    # Right stabilizers
    for stab in new_stabs_right:
        new_stab_to_circuit[stab.uuid] = next(
            syndrome_circuit.uuid
            for syndrome_circuit in block1.syndrome_circuits + block2.syndrome_circuits
            if syndrome_circuit.name == f"right-{stab.pauli.lower()}"
        )
    # Top stabilizers
    for stab in new_stabs_top:
        new_stab_to_circuit[stab.uuid] = next(
            syndrome_circuit.uuid
            for syndrome_circuit in block1.syndrome_circuits + block2.syndrome_circuits
            if syndrome_circuit.name == f"top-{stab.pauli.lower()}"
        )
    # Bottom stabilizers
    for stab in new_stabs_bottom:
        new_stab_to_circuit[stab.uuid] = next(
            syndrome_circuit.uuid
            for syndrome_circuit in block1.syndrome_circuits + block2.syndrome_circuits
            if syndrome_circuit.name == f"bottom-{stab.pauli.lower()}"
        )
    # Bulk stabilizers
    for stab in new_bulk_stabilizers:
        new_stab_to_circuit[stab.uuid] = next(
            syndrome_circuit.uuid
            for syndrome_circuit in block1.syndrome_circuits + block2.syndrome_circuits
            if syndrome_circuit.name == stab.pauli.lower()
        )

    return new_stab_to_circuit


def create_merge_2_body_stabilizers(
    blocks: tuple[RotatedSurfaceCode, RotatedSurfaceCode],
    merge_orientation: Orientation,
    merge_distance: int,
    filling_upper_left_qubit: tuple[int, int],
    pauli: str,
) -> tuple[list[Stabilizer], list[Stabilizer]]:
    """
    Create the new 2-body stabilizers located at the merged boundaries.

    Parameters
    ----------
    blocks : tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        blocks to merge.
    merge_orientation : Orientation
        Orientation of the merge.
    merge_distance : int
        Distance between the two blocks in the direction of the merge. A distance m
        means that there are m-1 rows/columns of data qubits between the two blocks
        to merge.
    filling_upper_left_qubit : tuple[int, int]
        Upper left qubit of the filling region.
    pauli : str
        Pauli string of the new stabilizers.

    Returns
    -------
    tuple[list[Stabilizer], list[Stabilizer]]
        New left and right boundary stabilizers if the merge is horizontal, new top and
        bottom boundary stabilizers if the merge is vertical.
    """
    block1, _ = blocks
    is_merge_horizontal = merge_orientation == Orientation.HORIZONTAL

    # Index in block.size corresponding to the merge direction, 0 for horizontal, 1 for
    # vertical
    size_index = 0 if is_merge_horizontal else 1

    # The two boundaries merged both have a 2 body stabilizer in the first row/column if
    # the distance normal to the merge direction is odd, e.g. for a 3x3 block,
    # weight_2_stab_is_first_row=True and a vertical merge, the additional left boundary
    # will have a stabilizer in the first row but the right boundary does not.
    boundaries_have_same_weight_2_stab_is_first = block1.size[not size_index] % 2 == 0

    if is_merge_horizontal:
        block1_weight_2_stab_is_first_column = not block1.weight_2_stab_is_first_row
        # If the size of block1 is even, the filling's weight_2_stab_is_first_column is
        # the same as block1's else it's opposite
        filling_weight_2_stab_is_first_top_column = (
            block1_weight_2_stab_is_first_column ^ ((block1.size[size_index] % 2) == 0)
        )
        filling_weight_2_stab_is_first_bot_column = (
            filling_weight_2_stab_is_first_top_column
            if boundaries_have_same_weight_2_stab_is_first
            else not filling_weight_2_stab_is_first_top_column
        )
        n_stabs_top = merge_distance // 2 + (
            filling_weight_2_stab_is_first_top_column and (merge_distance % 2)
        )
        n_stabs_bottom = merge_distance // 2 + (
            filling_weight_2_stab_is_first_bot_column and (merge_distance % 2)
        )
        new_stabs_top = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=pauli,
            initial_position=(
                filling_upper_left_qubit[0]
                + (not filling_weight_2_stab_is_first_top_column),
                filling_upper_left_qubit[1],
            ),
            num_stabs=n_stabs_top,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=False,
        )
        new_stabs_bottom = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=pauli,
            initial_position=(
                filling_upper_left_qubit[0]
                + (not filling_weight_2_stab_is_first_bot_column),
                filling_upper_left_qubit[1] + block1.size[1] - 1,
            ),
            num_stabs=n_stabs_bottom,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=True,
        )

        return new_stabs_top, new_stabs_bottom

    # else
    # If the size of block 1 is even, the filling's weight_2_stab_is_first_row is
    # the same as block 1 else it's opposite
    filling_weight_2_stab_is_first_left_row = block1.weight_2_stab_is_first_row ^ (
        (block1.size[size_index] % 2) == 0
    )
    filling_weight_2_stab_is_first_right_row = (
        filling_weight_2_stab_is_first_left_row
        if boundaries_have_same_weight_2_stab_is_first
        else not filling_weight_2_stab_is_first_left_row
    )
    n_stabs_left = merge_distance // 2 + (
        filling_weight_2_stab_is_first_left_row and (merge_distance % 2)
    )
    n_stabs_right = merge_distance // 2 + (
        filling_weight_2_stab_is_first_right_row and (merge_distance % 2)
    )
    new_stabs_left = RotatedSurfaceCode.generate_weight2_stabs(
        pauli=pauli,
        initial_position=(
            filling_upper_left_qubit[0],
            filling_upper_left_qubit[1] + (not filling_weight_2_stab_is_first_left_row),
        ),
        num_stabs=n_stabs_left,
        orientation=Orientation.VERTICAL,
        is_bottom_or_right=False,
    )
    new_stabs_right = RotatedSurfaceCode.generate_weight2_stabs(
        pauli=pauli,
        initial_position=(
            filling_upper_left_qubit[0] + block1.size[0] - 1,
            filling_upper_left_qubit[1]
            + (not filling_weight_2_stab_is_first_right_row),
        ),
        num_stabs=n_stabs_right,
        orientation=Orientation.VERTICAL,
        is_bottom_or_right=True,
    )

    return new_stabs_left, new_stabs_right


def merge_stabilizers(  # pylint: disable=too-many-locals
    blocks: tuple[RotatedSurfaceCode, RotatedSurfaceCode],
    merge_is_horizontal: bool,
) -> tuple[list[Stabilizer], list[Stabilizer], list[Stabilizer], dict[str, str]]:
    """Merge the two initial blocks stabilizers into a single lists of stabilizers, the
    list of old stabilizers that are gonna be replaced and the new stabilizers that will
    replace them.

    Parameters
    ----------
    blocks : tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        Blocks to merge. By construction blocks[0] should be the left or top block and
        blocks[1] should be the right or bottom block.
    merge_is_horizontal : bool
        True if the merge is horizontal, False if the merge is vertical.

    Returns
    -------
    tuple[list[Stabilizer], list[Stabilizer], list[Stabilizer], dict[str, str]]
        List of stabilizers forming the new block, list of old stabilizers to be
        replaced, list of new stabilizers which will replace them and map from
        stabilizer to the syndrome circuits.
    """
    # C) - STABILIZERS
    #    C.1) Create the new 4-body stabilizers located between the two initial blocks
    block1, block2 = blocks
    weight_4_x_schedule = block1.weight_4_x_schedule
    weight_4_z_schedule = block1.weight_4_z_schedule

    # Find the qubit located at the upper left corner of the filling
    # block1 is always the left or top block and block2 is always the right or bottom
    # block
    boundary_block_1 = (
        block1.boundary_qubits("right")
        if merge_is_horizontal
        else block1.boundary_qubits("bottom")
    )
    boundary_block_2 = (
        block2.boundary_qubits("left")
        if merge_is_horizontal
        else block2.boundary_qubits("top")
    )

    # The "filling" is the set of new 4 body stabilizers and 2-body stabilizers
    upper_left_qubit_filling = min(boundary_block_1, key=lambda x: x[0] + x[1])
    dx = (
        block2.upper_left_qubit[0] - upper_left_qubit_filling[0] + 1
        if merge_is_horizontal
        else block1.size[0]
    )
    dz = (
        block2.upper_left_qubit[1] - upper_left_qubit_filling[1] + 1
        if not merge_is_horizontal
        else block1.size[1]
    )

    adjacent_stab = next(
        stab
        for stab in block1.stabilizers
        if upper_left_qubit_filling in stab.data_qubits and len(stab.pauli) == 4
    )
    upper_left_4_body_pauli = "XXXX" if adjacent_stab.pauli[0] == "Z" else "ZZZZ"
    new_upleft_4_body_stabs = RotatedSurfaceCode.generate_weight4_stabs(
        pauli=upper_left_4_body_pauli,
        schedule=(
            weight_4_x_schedule
            if upper_left_4_body_pauli == "XXXX"
            else weight_4_z_schedule
        ),
        start_in_top_left_corner=True,
        dx=dx,
        dz=dz,
        initial_position=upper_left_qubit_filling,
    )
    new_rest_4_body_stabs = RotatedSurfaceCode.generate_weight4_stabs(
        pauli="XXXX" if upper_left_4_body_pauli == "ZZZZ" else "ZZZZ",
        schedule=(
            weight_4_x_schedule
            if upper_left_4_body_pauli != "XXXX"
            else weight_4_z_schedule
        ),
        start_in_top_left_corner=False,
        dx=dx,
        dz=dz,
        initial_position=upper_left_qubit_filling,
    )
    new_4_body_stabs = new_upleft_4_body_stabs + new_rest_4_body_stabs

    #    C.2) Create the new 2-body stabilizers
    stab_left_right_is_x = (
        block1.x_boundary == Orientation.HORIZONTAL
    )  # The logical operators are aligned

    stab_left_right = "XX" if stab_left_right_is_x else "ZZ"
    stab_top_bottom = "ZZ" if stab_left_right_is_x else "XX"

    if merge_is_horizontal:
        stabs_top, stabs_bottom = create_merge_2_body_stabilizers(
            blocks=blocks,
            merge_orientation=Orientation.HORIZONTAL,
            merge_distance=dx - 1,  # dx is the number of qubits used in the filling
            filling_upper_left_qubit=upper_left_qubit_filling,
            pauli=stab_top_bottom,
        )
        stabs_left, stabs_right = [], []
    else:
        stabs_top, stabs_bottom = [], []
        stabs_left, stabs_right = create_merge_2_body_stabilizers(
            blocks=blocks,
            merge_orientation=Orientation.VERTICAL,
            merge_distance=dz - 1,  # dz is the number of qubits used in the filling
            filling_upper_left_qubit=upper_left_qubit_filling,
            pauli=stab_left_right,
        )

    new_2_body_stabs = stabs_top + stabs_bottom + stabs_left + stabs_right

    #    C.3) Find the 2-body stabilizers which weight should be increased
    old_stabs_to_lengthen = [
        stab
        for stab in block1.stabilizers + block2.stabilizers
        if set(stab.data_qubits).issubset(set(boundary_block_1 + boundary_block_2))
        and len(stab.data_qubits) == 2
    ]
    #    C.4) Find the associated 4-body stabilizers
    #    There is a 1-to-1 correspondence between old_stabs_to_lengthen and
    #    new_stabs_increased_weight
    # NOTE: the order in the product matters, we want to match the order
    # old_stabs_to_lengthen to new_stabs_increased_weight
    new_stabs_increased_weight = [
        stab
        for (old_stab, stab) in product(old_stabs_to_lengthen, new_4_body_stabs)
        if set(old_stab.data_qubits).issubset(set(stab.data_qubits))
        and old_stab.pauli in stab.pauli
    ]

    #    C.5) Create the new `stabilizer_to_circuit` mapping
    new_stab_to_circ = find_merge_stabilizer_to_circuit_mappings(
        blocks,
        stabs_left,
        stabs_right,
        stabs_top,
        stabs_bottom,
        new_4_body_stabs,
        old_stabs_to_lengthen,
    )

    #    C.6) Create a single set of stabilizers for the new block
    new_stabilizers = new_4_body_stabs + new_2_body_stabs

    new_block_stabilizers = [
        stab
        for stab in block1.stabilizers + block2.stabilizers
        if stab not in old_stabs_to_lengthen
    ] + new_stabilizers

    return (
        new_block_stabilizers,
        old_stabs_to_lengthen,
        new_stabs_increased_weight,
        new_stab_to_circ,
    )


def find_data_qubits_between(
    qubit1: tuple[int, ...], qubit2: tuple[int, ...]
) -> list[tuple[int, ...]]:
    """
    Generate all qubits between two given qubits coordinates (not inclusive).

    Parameters
    ----------
    qubit1: tuple[int, ...]:
        The first qubit (x1, y1, ...).
    qubit2: tuple[int, ...]:
        The second qubit (x2, y2, ...).

    Returns
    -------
    list[tuple[int, ...]]
        A list of qubits between the two qubits.
    """
    x1, y1 = qubit1[:2]
    x2, y2 = qubit2[:2]

    if x1 == x2:
        # Vertical alignment
        y_start, y_end = sorted([y1, y2])
        qubits_between = [(x1, y, 0) for y in range(y_start + 1, y_end)]
    elif y1 == y2:
        # Horizontal alignment
        x_start, x_end = sorted([x1, x2])
        qubits_between = [(x, y1, 0) for x in range(x_start + 1, x_end)]
    else:
        raise ValueError("The qubits are not aligned horizontally or vertically.")

    return qubits_between


def merge_logical_operators(
    interpretation_step: InterpretationStep,
    blocks: tuple[RotatedSurfaceCode, RotatedSurfaceCode],
    qubits_to_reset: tuple[tuple[int, ...]],
    merge_is_horizontal: bool,
) -> tuple[
    InterpretationStep, PauliOperator, tuple[Cbit, ...], PauliOperator, tuple[Cbit, ...]
]:
    """Generate the new logical operators from the two blocks to merge. Note that they
    will be pushed to the upper left corner of the new block, similarly to the default
    choice of logical operators. This function also retrieves the lates measurements of
    the equivalent stabilizers and returns them.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks to merge.
    blocks : tuple[RotatedSurfaceCode, RotatedSurfaceCode]
        Initial blocks to merge. By convention, the first block must be the one on the
        left or top. The second block must be the one on the right or bottom
        respectively.
    qubits_to_reset : tuple[tuple[int, ...]]
        Qubits to measure between the two blocks.
    merge_is_horizontal : bool
        Whether the merge is horizontal or vertical.

    Returns
    -------
    tuple[
        InterpretationStep,
        PauliOperator,
        tuple[Cbit, ...],
        PauliOperator,
        tuple[Cbit, ...]
    ]
        The updated interpretation step, the new logical X operator, the measurements
        coming from equivalent X stabilizers, the new logical Z operator and the
        measurements coming from equivalent Z stabilizers.
    """
    # D) LOGICAL OPERATORS
    block1, block2 = blocks
    x_log_is_horizontal = block1.x_boundary == Orientation.HORIZONTAL

    # The logical operator aligned with the merge will be modified
    if x_log_is_horizontal == merge_is_horizontal:
        pauli_to_be_merged = "X"
        initial_log_ops_to_be_merged = (
            block1.logical_x_operators[0],
            block2.logical_x_operators[0],
        )
        log_ops_untouched = (
            block1.logical_z_operators[0],
            block2.logical_z_operators[0],
        )
    else:
        pauli_to_be_merged = "Z"
        initial_log_ops_to_be_merged = (
            block1.logical_z_operators[0],
            block2.logical_z_operators[0],
        )
        log_ops_untouched = (
            block1.logical_x_operators[0],
            block2.logical_x_operators[0],
        )

    #    D.1) Align the 2 initial operators with stabilizer products
    aligned_logical_block_1, extra_stabs1 = (
        block1.get_shifted_equivalent_logical_operator(
            initial_log_ops_to_be_merged[0],
            block1.upper_left_qubit,
        )
    )
    aligned_logical_block_2, extra_stabs2 = (
        block2.get_shifted_equivalent_logical_operator(
            initial_log_ops_to_be_merged[1],
            block2.upper_left_qubit,
        )
    )

    #    D.2) Retrieve the cbits coming from the latest equivalent stabilizer
    #    measurements.
    cbits_block_1 = interpretation_step.retrieve_cbits_from_stabilizers(
        extra_stabs1, block1
    )
    cbits_block_2 = interpretation_step.retrieve_cbits_from_stabilizers(
        extra_stabs2, block2
    )

    #    D.3) Create new logical operators using qubits to measure
    # We assume the logical operators are on a line and located between the end of the
    # first logical operator and the start of the second one
    qubits_in_between = find_data_qubits_between(
        max(aligned_logical_block_1.data_qubits, key=lambda x: x[0] + x[1]),
        min(aligned_logical_block_2.data_qubits, key=lambda x: x[0] + x[1]),
    )
    if not set(qubits_in_between).issubset(set(qubits_to_reset)):
        raise RuntimeError("The qubits chosen are not included in the reset qubits.")

    qubits_new_logical = (
        aligned_logical_block_1.data_qubits
        + aligned_logical_block_2.data_qubits
        + tuple(qubits_in_between)
    )

    new_logical = PauliOperator(
        pauli=pauli_to_be_merged * len(qubits_new_logical),
        data_qubits=qubits_new_logical,
    )
    untouched_logical = log_ops_untouched[0]
    new_log_x, new_log_z = (
        (new_logical, untouched_logical)
        if new_logical.pauli[0] == "X"
        else (untouched_logical, new_logical)
    )

    #    D.4) Update `logical_x/z_evolution`
    if pauli_to_be_merged == "X":
        interpretation_step.logical_x_evolution[new_logical.uuid] = (
            initial_log_ops_to_be_merged[0].uuid,
            initial_log_ops_to_be_merged[1].uuid,
        ) + tuple(stab.uuid for stab in extra_stabs1 + extra_stabs2)
        cbits_x = cbits_block_1 + cbits_block_2
        cbits_z = ()
    else:
        interpretation_step.logical_z_evolution[new_logical.uuid] = (
            initial_log_ops_to_be_merged[0].uuid,
            initial_log_ops_to_be_merged[1].uuid,
        ) + tuple(stab.uuid for stab in extra_stabs1 + extra_stabs2)
        cbits_x = ()
        cbits_z = cbits_block_1 + cbits_block_2

    return interpretation_step, new_log_x, cbits_x, new_log_z, cbits_z


def create_syndromes(  # pylint: disable= too-many-arguments
    interpretation_step: InterpretationStep,
    qubits_to_reset: tuple[tuple[int, int, int], ...],
    reset_type: str,
    new_stabilizers: tuple[Stabilizer, ...],
    new_stabs_increased_weight: tuple[Stabilizer, ...],
    merged_block: Block,
) -> tuple[Syndrome, ...]:
    """Creates the new syndromes for the stabilizers that are reset.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step to which we add the new syndromes.
    qubits_to_reset : tuple[tuple[int, int, int], ...]
        Tuple of qubits that are reset during the merge
    reset_type:
        Pauli type of the reset, this gives information on which stabilizer is reset in
        a deterministic state.
    new_stabilizers : tuple[Stabilizer, ...]
        Stabilizers created by the merge operation.
    new_stabs_increased_weight:
        Stabilizers that grew from a previous 2-body of the initial blocks.
    merged_block : Block
       The newly merged block.

    Returns
    -------
    tuple[Syndrome, ...]
        Reference syndromes of the reset stabilizers.
    """
    reset_stabilizers = tuple(
        stab
        for stab in new_stabilizers
        if (
            all(q in qubits_to_reset for q in stab.data_qubits)
            and (stab not in new_stabs_increased_weight)
            and set(stab.pauli) == {reset_type}
        )
    )
    syndromes = generate_syndromes(
        interpretation_step=interpretation_step,
        stabilizers=reset_stabilizers,
        block=merged_block,
        stab_measurements=tuple(() for _ in reset_stabilizers),
    )
    return syndromes


def merge(  # pylint: disable=line-too-long, too-many-locals
    interpretation_step: InterpretationStep,
    operation: Merge,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Merge two blocks specified in the Merge operation.

    The algorithm is the following:

    - A.) DATA QUBITS
    
        - A.1) Find data qubits to be measured in between the two blocks, they will be \
        merged into a single new block
        
    - B.) CIRCUIT
    
        - B.1) Create classical channels for all data qubit measurements
        - B.2) Create a measurement circuit for every measured data qubit
        - B.3) Append the measurement circuits to the InterpretationStep circuit. \
        If needed, apply a basis change
            
    - C.) - STABILIZERS
    
        - C.1) Create the new 4-body stabilizers located between the two initial blocks
        - C.2) Create the new 2-body stabilizers located at the merged boundaries
        - C.3) Find the 2-body stabilizers which weight should be increased
        - C.4) Find the associated 4-body stabilizers
        - C.5) Create the new ``stabilizer_to_circuit`` mapping
        - C.6) Create a single set of stabilizers for the new block
        - C.7) Update ``stabilizer_evolution`` and ``stabilizer_updates`` for the \
        stabilizers which have been increased in weight
        
    - D.) LOGICAL OPERATORS
    
        - D.1) Align the 2 initial operators with stabilizer products
        - D.2) Retrieve the cbits coming from the latest equivalent stabilizer \
        measurements.
        - D.3) Create new logical operators using qubits to measure
        - D.4) Update ``logical_x/z_evolution``
        - D.5) Update ``logical_x/z_updates``
        
    - E.) NEW BLOCK AND NEW INTERPRETATION STEP
    
        - E.1) Create the new block
        - E.2) Update the block history
        
    - F.) SYNDROMES
    
        - F.1) Generate syndromes for newly created stabilizers

    - G) JOINT OBSERVABLES

        - G.1) Obtain the cbits for joint observable and store them in the
               interpretation_step

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    operation : Merge
        Descriptor of the merge operation to perform.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Modified interpretation step after the merge operation.
    """
    # Consistency checks
    block1, block2 = merge_consistency_check(interpretation_step, operation)

    # A) - DATA QUBITS
    #    Find data qubits to be measured in between the two blocks, they will be
    #    merged into a single new block
    link_vector = [
        coord2 - coord1
        for coord1, coord2 in zip(
            block1.upper_left_qubit, block2.upper_left_qubit, strict=True
        )
    ]
    index_nonzero = next((i for i, coord in enumerate(link_vector) if coord != 0))
    # Ensure that block1 is left or top and block2 is right or bottom
    if link_vector[index_nonzero] < 0:
        block1, block2 = block2, block1

    merge_is_horizontal = not bool(index_nonzero)

    if merge_is_horizontal:
        boundary_qubits_block1 = block1.boundary_qubits("right")
        _ = block2.boundary_qubits("left")
        distance = block2.upper_left_qubit[0] - (
            block1.upper_left_qubit[0] + block1.size[0]
        )
    else:
        boundary_qubits_block1 = block1.boundary_qubits("bottom")
        _ = block2.boundary_qubits("top")
        distance = block2.upper_left_qubit[1] - (
            block1.upper_left_qubit[1] + block1.size[1]
        )

    # Create a square of qubits to reset between the two blocks
    qubits_to_reset = [
        (
            qubit[0] + i * merge_is_horizontal,
            qubit[1] + i * (not merge_is_horizontal),
            0,
        )
        for qubit in boundary_qubits_block1
        for i in range(1, distance + 1)
    ]

    # B) - CIRCUIT
    #    B.1) Create classical channels for all data qubit measurements
    #    B.2) Create a measurement circuit for every measured data qubit
    boundary_type = (
        block1.boundary_type("right")
        if merge_is_horizontal
        else block1.boundary_type("bottom")
    )
    merge_circuit = create_merge_circuit(
        interpretation_step,
        (block1, block2),
        operation,
        qubits_to_reset,
        boundary_type,
    )
    #    B.3) Append the measurement circuits to the InterpretationStep circuit
    #       If needed, apply a basis change
    interpretation_step.append_circuit_MUT(merge_circuit, same_timeslice)

    # C) - STABILIZERS
    #    C.1) Create the new 4-body stabilizers located between the two initial blocks
    #    C.2) Create the new 2-body stabilizers located at the merged boundaries
    #    C.3) Find the 2-body stabilizers which weight should be increased
    #    C.4) Find the associated 4-body stabilizers
    #    C.5) Create the new `stabilizer_to_circuit` mapping
    #    C.6) Create a single set of stabilizers for the new block
    (
        new_block_stabilizers,
        old_stabs_to_lengthen,
        new_stabs_increased_weight,
        new_stabilizer_to_circuit,
    ) = merge_stabilizers((block1, block2), merge_is_horizontal)

    #    C.7) Update `stabilizer_evolution` and `stabilizer_updates` for the
    #         stabilizers which have been increased in weight
    # Stabilizer evolution: Update the stabilizer evolution dictionary
    stab_map_weight2_to_weight4 = {
        new_stab.uuid: (stab.uuid,)
        for new_stab, stab in zip(
            new_stabs_increased_weight, old_stabs_to_lengthen, strict=True
        )
    }
    interpretation_step.stabilizer_evolution.update(stab_map_weight2_to_weight4)
    # Stabilizer updates: Since we use a reset, there is no update to the stabilizers

    # D) LOGICAL OPERATORS
    #    D.1) Align the 2 initial operators with stabilizer products
    #    D.2) Retrieve the cbits coming from the latest equivalent stabilizer
    #         measurements.
    #    D.3) Create new logical operators using qubits to measure
    #    D.4) Update `logical_x/z_evolution`
    interpretation_step, new_log_x, cbits_x, new_log_z, cbits_z = (
        merge_logical_operators(
            interpretation_step,
            (block1, block2),
            qubits_to_reset,
            merge_is_horizontal,
        )
    )

    #    D.5) Update `logical_x/z_updates`
    # Inherit the updates from previous operators (only if the operator has changed),
    # adding the Cbits coming from moving logical operators
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="X",
        logical_operator_id=new_log_x.uuid,
        new_updates=cbits_x,
        inherit_updates=(
            new_log_x.uuid
            not in (
                block1.logical_x_operators[0].uuid,
                block2.logical_x_operators[0].uuid,
            )
        ),
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="Z",
        logical_operator_id=new_log_z.uuid,
        new_updates=cbits_z,
        inherit_updates=(
            new_log_z.uuid
            not in (
                block1.logical_z_operators[0].uuid,
                block2.logical_z_operators[0].uuid,
            )
        ),
    )

    # E) NEW BLOCK AND NEW INTERPRETATION STEP
    #    E.1) Create the new block
    merged_block = RotatedSurfaceCode(
        stabilizers=list(new_block_stabilizers),
        logical_x_operators=[new_log_x],
        logical_z_operators=[new_log_z],
        unique_label=operation.output_block_name,
        syndrome_circuits=block1.syndrome_circuits,
        stabilizer_to_circuit=new_stabilizer_to_circuit,
        skip_validation=not debug_mode,
    )
    #    E.2) Update the block history
    # Update only the blocks that are involved in the merge
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(merged_block,),
        old_blocks=(block1, block2),
    )

    # F) SYNDROMES
    #    F.1) Generate syndromes for newly created stabilizers
    merge_syndromes = create_syndromes(
        interpretation_step=interpretation_step,
        qubits_to_reset=qubits_to_reset,
        reset_type=("X" if boundary_type == "Z" else "Z"),
        new_stabilizers=new_block_stabilizers,
        new_stabs_increased_weight=new_stabs_increased_weight,
        merged_block=merged_block,
    )
    interpretation_step.append_syndromes_MUT(merge_syndromes)

    # G) JOINT OBSERVABLE

    # Measure blocks once to obtain the cbits for the stabilizers in between
    interpretation_step = measureblocksyndromes(
        interpretation_step,
        MeasureBlockSyndromes(merged_block.unique_label, 1),
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    new_log_op, old_log_op_2, log_updates = (
        (
            merged_block.logical_x_operators[0],
            block2.logical_x_operators[0],
            interpretation_step.logical_x_operator_updates,
        )
        if boundary_type == "X"
        else (
            merged_block.logical_z_operators[0],
            block2.logical_z_operators[0],
            interpretation_step.logical_z_operator_updates,
        )
    )

    # Obtain stabilizers between the logical operators of block1 and block2.
    # Note that by default, the new logical operator is equal to that of block1.
    _, stabs_between = merged_block.get_shifted_equivalent_logical_operator(
        new_log_op, min(old_log_op_2.data_qubits, key=lambda x: x[0] + x[1])
    )

    # Retrieve cbits associated with `stabs_between` from the latest stabilizer
    # measurements.
    stabs_between_cbits = interpretation_step.retrieve_cbits_from_stabilizers(
        stabs_between, merged_block
    )

    log_corrections = log_updates.get(new_log_op.uuid, ()) + log_updates.get(
        old_log_op_2.uuid, ()
    )

    joint_observable = stabs_between_cbits + log_corrections

    interpretation_step.logical_measurements[
        LogicalMeasurement(
            blocks=(block1.unique_label, block2.unique_label),
            observable=boundary_type * 2,
        )
    ] = joint_observable

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-lines
from __future__ import annotations
from enum import Enum

from loom.eka import Circuit, Stabilizer, SyndromeCircuit, PauliOperator
from loom.eka.utilities import Direction, DiagonalDirection
from ..utilities.enums import FourBodySchedule
from ..code_factory.rotated_surface_code import RotatedSurfaceCode


def add_vector(
    qubit: tuple[int, int, int], direction: Direction, length: int
) -> tuple[int, int, int]:
    """Add a direction vector to a qubit coordinate scaled by length. Used to
    calculate the new position of a corner after moving it."""
    direction_vect = direction.to_vector()
    qubit = list(qubit)
    for i, coord in enumerate(direction_vect):
        qubit[i] += coord * length
    return tuple(qubit)


def generate_syndrome_extraction_circuits(
    rsc_block: RotatedSurfaceCode, starting_qubit_diagonal_direction: DiagonalDirection
) -> tuple[tuple[SyndromeCircuit, ...], dict[str, str]]:
    """
    Given a rotated surface code block, generate syndrome extraction circuits for
    each stabilizer based on the starting diagonal direction. The generated syndrome
    circuits should lead to a fault-tolerant behavior if used appropriately (with an
    appropriate `starting_qubit_diagonal_direction`).

    Parameters
    ----------
    rsc_block: RotatedSurfaceCode
        A rotated surface code block.
    starting_qubit_diagonal_direction: DiagonalDirection
        The diagonal direction of the starting data qubit for each syndrome extraction
        circuit.

    Returns
    -------
    tuple[tuple[SyndromeCircuit, ...], dict[str, str]]:
        A tuple containing:

        - A tuple of generated syndrome extraction circuits.
        - A dictionary mapping stabilizer UUIDs to their corresponding syndrome \
        circuit UUIDs.

    """
    # find stabilizer schedules
    (
        non_triangle_x_schedule,
        non_triangle_z_schedule,
        triangle_x_schedule,
        triangle_z_schedule,
    ) = find_schedules(rsc_block)

    # create syndrome circuits and new stabilizers with correct schedules
    (
        new_syndrome_circuits,
        new_stabilizer_to_circuit,
    ) = create_new_syndrome_circuits_with_known_schedules(
        rsc_block,
        non_triangle_x_schedule,
        non_triangle_z_schedule,
        triangle_x_schedule,
        triangle_z_schedule,
        starting_qubit_diagonal_direction,
    )

    return (
        new_syndrome_circuits,
        new_stabilizer_to_circuit,
    )


def find_schedules(rsc_block: RotatedSurfaceCode) -> tuple[
    FourBodySchedule,
    FourBodySchedule,
    FourBodySchedule | None,
    FourBodySchedule | None,
]:
    """
    Determine schedule for stabilizers of the block.

    For type 1 and 2 corner configuration (rectangle and U-shape), schedules can be
    inferred from boundary type of the short edge connecting two
    topological corners.

    Example::

        (type 1)
        1-------4
        |       |
        |       |
        |       |
        2-------3

        (type 2)
        1-------|
        |       |
        |       2
        |       |
        3-------4


    For type 3 corner configuration (L-shape), draw a diagonal line through the
    middle-edge topological corner and the geometric corner that is not occupied by
    a topological corner. This line divides the block into two parts, one triangle
    part and one non-triangle part. Return schedules for X and Z-type stabilizers in
    these parts.

    Example: the line is drawn through topological corner 2 and geometric corner x::

        (type 3)
        1-------x
        |       |
        2       |
        |       |
        3-------4

    For type 4 corner configuration (U-shape phase gate), draw a diagonal line through
    the middle-edge topological corner towards the direction of the angle topological
    corner. This line divides the block into two parts, one triangle part and one
    non-triangle part. Return schedules for X and Z-type stabilizers in these parts.

    Example: the line is drawn through topological corner 2 and geometric corner x::

        (type 4)
        1-------.
        |       |
        |       2
        |       |
        x       |
        3-------4

    The algorithm is the following:

    - A.) Find the boundary type of the short boundary connecting two topological \
        corners. For type 1, pick any short edge.
    - B.) Find schedules

        - B.1.) Boundary type is of opposite Pauli type to the logical operator \
        running along the boundary. \
        (e.g. boundary_type = "X" means logical Z running along the boundary)
        - B.2.) For type 1 (rectangle) and type 2 (U shape), the opposite short edge \
        has the same Pauli type. The schedule of stabilizers with the same Pauli \
        type as the short edge is determined so that the propagation of the \
        opposite Pauli type is perpendicular to the short edge. The schedule of \
        the-opposite-type stabilizers is the opposite schedule of the above \
        schedule.
        - B.3.) For type 3 (L shape), the opposite short edge has the opposite Pauli \
        type. Inside the triangle, the schedule for stabilizers with the same type \
        as the short edge is determined. Then the schedule for the other type \
        stabilizers inside the triangle is the opposite schedule. Outside the \
        triangle, the schedules are the opposite schedules of the same stabilizer type.
        - B.4.) For type 4 (U-shape phase gate), inside the triangle, the schedule for \
        stabilizers with the same type as the short edge is determined. Then the \
        schedule for the other type of stabilizers is the opposite schedule. \
        Outside the triangle, the schedules of both type stabilizers follow the \
        schedule for stabilizers with the same type as the short edge inside the \
        triangle.

    Returns
    -------
    FourBodySchedule:
        The schedule type of X-stabilizer in the non-triangle part (for 1,2,3, config).
    FourBodySchedule:
        The schedule type of Z-stabilizer in the non-triangle part (for 1,2,3, config).
    FourBodySchedule | None:
        The schedule type of X-stabilizer in the triangle part (for type 3 config).
        None for other type of config.
    FourBodySchedule | None:
        The schedule type of Z-stabilizer in the triangle part (for type 3 config).
        None for other type of config.
    """
    config, pivot_corners = rsc_block.config_and_pivot_corners
    is_horizontal = rsc_block.is_horizontal
    stabilizers = rsc_block.stabilizers
    if config == 0:
        triangle_x_schedule = triangle_z_schedule = None
        non_triangle_x_schedule = rsc_block.weight_4_x_schedule
        non_triangle_z_schedule = non_triangle_x_schedule.opposite_schedule()
        return (
            non_triangle_x_schedule,
            non_triangle_z_schedule,
            triangle_x_schedule,
            triangle_z_schedule,
        )

    short_end_corner = pivot_corners[3]  # short end corner
    long_edge_idx = 0 if is_horizontal else 1  # coordinate index along the long side

    # A) Find boundary type for the short-edge boundary passing through short_end_corner
    weight2_stabs_at_boundary = [
        stab
        for stab in stabilizers
        if len(stab.pauli) == 2
        and all(
            dqubit[long_edge_idx] == short_end_corner[long_edge_idx]
            for dqubit in stab.data_qubits
        )
    ]
    boundary_paulis = [stab.pauli for stab in weight2_stabs_at_boundary]
    if len(set(boundary_paulis)) != 1:
        raise ValueError(
            f"Accept only one type of stabilizers along short-edge boundary, "
            f"found {boundary_paulis}."
        )
    boundary_type = boundary_paulis[0][0]

    # B) Find schedules
    match config:
        case 1 | 2 | 4:
            if boundary_type == "Z":
                # logical X runs along the short edge
                non_triangle_x_schedule = (
                    FourBodySchedule.Z if is_horizontal else FourBodySchedule.N
                )
            else:
                # logical Z runs along the short edge
                non_triangle_x_schedule = (
                    FourBodySchedule.N if is_horizontal else FourBodySchedule.Z
                )
            non_triangle_z_schedule = non_triangle_x_schedule.opposite_schedule()
            triangle_x_schedule, triangle_z_schedule = None, None
        case 3:
            if boundary_type == "Z":
                # logical Z runs along the short edge of the triangle part
                triangle_x_schedule = (
                    FourBodySchedule.N if is_horizontal else FourBodySchedule.Z
                )
            else:
                # logical X runs along the short edge of the triangle part
                triangle_x_schedule = (
                    FourBodySchedule.Z if is_horizontal else FourBodySchedule.N
                )
            triangle_z_schedule = triangle_x_schedule.opposite_schedule()
            non_triangle_x_schedule = triangle_x_schedule.opposite_schedule()
            non_triangle_z_schedule = triangle_z_schedule.opposite_schedule()
        case _:
            raise ValueError(f"Unknown corner configuration {config}.")

    return (
        non_triangle_x_schedule,
        non_triangle_z_schedule,
        triangle_x_schedule,
        triangle_z_schedule,
    )


def find_stabilizer_position(
    rsc_block: RotatedSurfaceCode,
    config: int,
    pivot_corners: tuple[
        tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]
    ],
    stab: Stabilizer,
    is_horizontal: bool,
) -> tuple[bool, Direction | None]:
    # pylint: disable=duplicate-code
    """
    Find stabilizer position relative to the triangle partition
    (in type 3 configuration)

    Parameters
    ----------
    rsc_block: RotatedSurfaceCode
        The initial block
    config: int
        Corner configuration
    pivot_corners:
        tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]]
        Four topological corners
    stab: Stabilizer
        The input stabilizer
    is_horizontal: bool
        True if the long edge of the block is horizontal

    Returns
    -------
    tuple[bool, Direction | None]
        True if the stabilizer is inside triangular part (if any)
        Direction of the boundary if the input stabilizer is a weight-2 stabilizer

    """

    if config == 0:
        is_in_triangle = False
    else:
        long_end_corner, middle_corner, _, short_end_corner = pivot_corners
        long_edge_idx = (
            0 if is_horizontal else 1
        )  # coordinate index along the long side
        short_edge_idx = (
            1 if is_horizontal else 0
        )  # coordinate index along the short side

        def is_triangular_qubit(q) -> bool:
            """Determine if the input qubit is inside the triangular part
            (including the boundaries of the triangle). This is done by checking
            whether the distance along the long side of the block between the input
            qubit and the middle-edge corner is at least the distance along the
            short side between the same two qubits.
            """
            match config:
                case 1 | 2 | 4:
                    return False
                case 3:
                    dshort = (
                        middle_corner[short_edge_idx] - q[short_edge_idx]
                        if middle_corner[short_edge_idx]
                        > short_end_corner[short_edge_idx]
                        else q[short_edge_idx] - middle_corner[short_edge_idx]
                    )
                    dlong = (
                        middle_corner[long_edge_idx] - q[long_edge_idx]
                        if middle_corner[long_edge_idx] > long_end_corner[long_edge_idx]
                        else q[long_edge_idx] - middle_corner[long_edge_idx]
                    )
                    return 0 <= dshort <= dlong
                case _:
                    raise ValueError(f"Unknown corner configuration {config}.")

        is_in_triangle = all(is_triangular_qubit(dqubit) for dqubit in stab.data_qubits)

    # find boundary direction for weight-2 stabilizer
    boundary = None
    weight = len(stab.pauli)
    if weight == 2:
        qubit_directions = [
            set(
                direction
                for direction in Direction
                if dqubit in rsc_block.boundary_qubits(direction)
            )
            for dqubit in stab.data_qubits
        ]
        boundary = list(qubit_directions[0].intersection(qubit_directions[1]))[0]

    return is_in_triangle, boundary


def create_new_syndrome_circuits_with_known_schedules(
    # pylint: disable=too-many-statements, too-many-branches, too-many-locals
    rsc_block: RotatedSurfaceCode,
    non_triangle_x_schedule: FourBodySchedule,
    non_triangle_z_schedule: FourBodySchedule,
    triangle_x_schedule: FourBodySchedule | None,
    triangle_z_schedule: FourBodySchedule | None,
    starting_qubit_diag_directions: DiagonalDirection,
) -> tuple[tuple[SyndromeCircuit, ...], dict[str, str]]:
    """
    Create the syndrome extraction circuits for each stabilizer in the input block
    based on the schedules provided in the input. The data qubits of each stabilizer
    are ordered according to the schedule and the starting qubit direction.

    Parameters
    ----------
    rsc_block: RotatedSurfaceCode
        The rotated surface code block.
    non_triangle_x_schedule: FourBodySchedule
        Schedule for X-type stabilizers in the non-triangle part.
    non_triangle_z_schedule: FourBodySchedule
        Schedule for Z-type stabilizers in the non-triangle part.
    triangle_x_schedule: FourBodySchedule | None
        Schedule for X-type stabilizers in the triangle part. None if there is no
        triangle part.
    triangle_z_schedule: FourBodySchedule | None
        Schedule for Z-type stabilizers in the triangle part. None if there is no
        triangle part.
    starting_qubit_diag_directions: DiagonalDirection
        The diagonal direction of the starting data qubit for each syndrome extraction
        circuit.

    Returns
    -------
    tuple[tuple[SyndromeCircuit, ...], dict[str, str]]:
        A tuple containing:
        - A tuple of generated syndrome extraction circuits.
        - A dictionary mapping stabilizer UUIDs to their corresponding syndrome
        circuit UUIDs.
    """

    def find_data_qubits_order(
        stab: Stabilizer,
        starting_qubit_diag_direction: DiagonalDirection,
        schedule: FourBodySchedule,
        boundary_direction: Direction | None,
    ) -> tuple[tuple[int, ...], ...]:
        """
        Find correct data qubit order with respect to the input schedule and starting
        qubit direction.

        Parameters
        ----------
        stab: Stabilizer
            The input stabilizer
        starting_qubit_diag_direction: tuple[Direction, Direction]
            Direction of the starting qubit
        schedule: FourBodySchedule
            Schedule
        boundary_direction: Direction | None
            Boundary direction if the input stabilizer is a weight-2 stabilizer

        Returns
        -------
        tuple[tuple[int, ...], ...]
            Ordered data qubits
        """
        weight = len(stab.pauli)
        is_n_schedule = schedule == FourBodySchedule.N
        lambda_expressions = {
            Direction.TOP: lambda x: -x[1],
            Direction.BOTTOM: lambda x: x[1],
            Direction.LEFT: lambda x: -x[0],
            Direction.RIGHT: lambda x: x[0],
        }

        qubits = list(stab.data_qubits)
        if weight == 2:
            # add padding qubits
            boundary_vect = boundary_direction.to_vector()
            qubits += [
                (dq[0] + boundary_vect[0], dq[1] + boundary_vect[1], dq[2])
                for dq in qubits
            ]

        # determine the starting qubit coordinate
        for direction in starting_qubit_diag_direction.components:
            max_value = max(
                lambda_expressions[direction](dq) for dq in stab.data_qubits
            )
            qubits = [
                dq for dq in qubits if lambda_expressions[direction](dq) == max_value
            ]
        starting_qubit = qubits[0]

        # determine the other qubits
        diag_vect = starting_qubit_diag_direction.to_vector()
        data_qubits = [
            starting_qubit,
            (
                starting_qubit[0] - diag_vect[0] * (not is_n_schedule),
                starting_qubit[1] - diag_vect[1] * is_n_schedule,
                starting_qubit[2],
            ),
            (
                starting_qubit[0] - diag_vect[0] * is_n_schedule,
                starting_qubit[1] - diag_vect[1] * (not is_n_schedule),
                starting_qubit[2],
            ),
            (
                starting_qubit[0] - diag_vect[0],
                starting_qubit[1] - diag_vect[1],
                starting_qubit[2],
            ),
        ]
        data_qubits = tuple(dq for dq in data_qubits if dq in stab.data_qubits)

        return data_qubits

    is_horizontal = rsc_block.is_horizontal
    config, pivot_corners = rsc_block.config_and_pivot_corners

    if triangle_x_schedule is None or triangle_z_schedule is None:
        if config == 3:
            raise ValueError(
                "Triangle schedules must be provided for type-3 corner configuration."
            )

    syndrome_circuits = tuple()
    stabilizer_to_circuit = {}
    for stab in rsc_block.stabilizers:
        pauli_type = stab.pauli_type
        is_in_triangular, boundary_direction = find_stabilizer_position(
            rsc_block, config, pivot_corners, stab, is_horizontal
        )
        if is_in_triangular:
            schedule = triangle_x_schedule if pauli_type == "X" else triangle_z_schedule
        else:
            schedule = (
                non_triangle_x_schedule
                if pauli_type == "X"
                else non_triangle_z_schedule
            )

        # Generate syndrome circuit based on triangular and boundary
        weight = stab.weight
        match config:
            case 0 | 1:
                # for standard block, name it as convention
                if weight == 4:
                    syndrome_circuit_name = stab.pauli.lower()
                elif weight == 2:
                    syndrome_circuit_name = f"{boundary_direction}-{stab.pauli.lower()}"
                else:
                    # Unsupported weight
                    continue
            case 2 | 3 | 4:
                # for non-standard block, name it randomly
                syndrome_circuit_name = (
                    f"{'triangular' if is_in_triangular else 'non_triangular'}-"
                    f"{boundary_direction if boundary_direction else 'bulk'}"
                    f"-{str(pauli_type) * len(stab.pauli)}"
                )
            case _:
                raise ValueError(f"Unknown corner configuration {config}.")

        padding = None
        if weight == 2:
            # find padding for weight-2 stabilizer
            padding_boundary = boundary_direction
            for direction in starting_qubit_diag_directions.components:
                # Padding function in RotatedSurfaceCode is written only for
                # TOP-RIGHT starting direction .
                # Infer padding for other starting directions.
                if direction not in [Direction.TOP, Direction.RIGHT]:
                    padding_boundary = padding_boundary.mirror_across_orientation(
                        direction.to_orientation().perpendicular()
                    )
            padding = RotatedSurfaceCode.find_padding(padding_boundary, schedule)

        syndrome_circuit = RotatedSurfaceCode.generate_syndrome_circuit(
            stab.pauli, padding, syndrome_circuit_name
        )
        # This syndrome circuit doesn't take into consideration that the order of data
        # qubits may not match the order with which they need to be entangled.
        # As a result, we are going to rearrange the channel order so that the data
        # qubits interact with the ancilla in the appropriate order

        # Rearrange the quantum channels in the Circuit part
        ordered_data_qubits = find_data_qubits_order(
            stab, starting_qubit_diag_directions, schedule, boundary_direction
        )
        ordered_data_qubit_indices = [
            ordered_data_qubits.index(dq) for dq in stab.data_qubits
        ]
        old_circuit = syndrome_circuit.circuit
        # Order the channels by first ordering the data qubits, and then append the rest
        # of the channels as they are
        channels_ordered = [
            old_circuit.channels[idx] for idx in ordered_data_qubit_indices
        ] + list(old_circuit.channels[len(ordered_data_qubit_indices) :])
        new_circuit = Circuit(old_circuit.name, old_circuit.circuit, channels_ordered)
        # Reconstruct the syndrome circuit using the new circuit with correct data qubit
        # order
        syndrome_circuit = SyndromeCircuit(
            syndrome_circuit.pauli, syndrome_circuit.name, new_circuit
        )

        # Avoid duplicate syndrome circuits
        same_syndrome_circuit = next(
            (circ for circ in syndrome_circuits if circ == syndrome_circuit), None
        )
        if same_syndrome_circuit is not None:
            syndrome_circuit = same_syndrome_circuit
        else:
            syndrome_circuits += (syndrome_circuit,)
        # Associate the stabilizer to the syndrome circuit
        stabilizer_to_circuit.update({stab.uuid: syndrome_circuit.uuid})

    syndrome_circuits = tuple(set(syndrome_circuits))

    return (
        syndrome_circuits,
        stabilizer_to_circuit,
    )


def direction_to_coord(  # pylint: disable=inconsistent-return-statements
    direction: Direction | DiagonalDirection, sublattice_index: int = None
) -> tuple[int, int, int]:
    """
    Converts a direction to a coordinate vector. This assumes the coordinates are in
    3-dimensions with the last coordinate being a sub-lattice index. The sub-lattice
    index of the current coordinate is also required for diagonal movements if the qubit
    is moved to a different sub-lattice.

    Parameters
    ----------
    direction: Direction | DiagonalDirection
        The direction in which the block is to be moved.
    sublattice_index: Optional[int]
        The sub-lattice index of the current coordinate. This should be defined for
        diagonal movements where the qubit is moved to a different sub-lattice.
        If it's a diagonal movement but on the same sub-lattice, this should be None.

    Returns
    -------
    tuple[int, int, int]
        The coordinate vector representing the direction in which the block is to be
        moved.

    Raises
    ------
    ValueError
        If the direction is not of type Direction or set of Directions of size 2.
        If the sub-lattice index is not an integer or None for diagonal movements.
        If the direction is not a valid direction.
    """
    # Direction vector for the 4 directions in the same sub-lattice.
    direction_vector = {
        Direction.TOP: (0, -1, 0),
        Direction.RIGHT: (1, 0, 0),
        Direction.LEFT: (-1, 0, 0),
        Direction.BOTTOM: (0, 1, 0),
    }
    # Sets are not hashable. Frozensets are.
    # The diagonal vector for moving within the same sub-lattice.
    diagonal_vector = {
        DiagonalDirection.TOP_RIGHT: (1, -1, 0),
        DiagonalDirection.TOP_LEFT: (-1, -1, 0),
        DiagonalDirection.BOTTOM_RIGHT: (1, 1, 0),
        DiagonalDirection.BOTTOM_LEFT: (-1, 1, 0),
    }
    # The diagonal vector for moving from sub-lattice 0 to sub-lattice 1.
    diagonal_vector_0 = {
        DiagonalDirection.TOP_RIGHT: (1, 0, 1),
        DiagonalDirection.TOP_LEFT: (0, 0, 1),
        DiagonalDirection.BOTTOM_RIGHT: (1, 1, 1),
        DiagonalDirection.BOTTOM_LEFT: (0, 1, 1),
    }
    # The diagonal vector for moving from sub-lattice 1 to sub-lattice 0.
    diagonal_vector_1 = {
        DiagonalDirection.TOP_RIGHT: (0, -1, -1),
        DiagonalDirection.TOP_LEFT: (-1, -1, -1),
        DiagonalDirection.BOTTOM_RIGHT: (0, 0, -1),
        DiagonalDirection.BOTTOM_LEFT: (-1, 0, -1),
    }

    if isinstance(direction, Direction):
        return direction_vector[direction]
    if isinstance(direction, DiagonalDirection):
        if not (isinstance(sublattice_index, int) or sublattice_index is None):
            raise ValueError(
                "Sub-lattice index must be an integer or None for diagonal movements."
            )
        if sublattice_index == 0:
            return diagonal_vector_0[direction]
        if sublattice_index == 1:
            return diagonal_vector_1[direction]
        if sublattice_index is None:
            return diagonal_vector[direction]
    else:
        raise ValueError(
            "Invalid direction. direction must be type Direction or set of Direction"
        )


def update_qubit_coords(
    input_qubit_coords: tuple[tuple[int, int, int]],
    direction: Direction | set[Direction],
) -> tuple[int, int, int]:
    """
    This function updates the coordinates of the qubits based on the direction of
    movement.
    """
    updated_qubit_coords = tuple()
    for each_qubit_coord in input_qubit_coords:
        updated_qubit_coords += (
            tuple(
                coord + dir
                for coord, dir in zip(
                    each_qubit_coord,
                    direction_to_coord(direction, each_qubit_coord[-1]),
                    strict=True,
                )
            ),
        )

    return updated_qubit_coords


def shift_block_towards_direction(
    rsc_block: RotatedSurfaceCode,
    direction: Direction | DiagonalDirection,
    debug_mode,
) -> tuple[
    RotatedSurfaceCode,
    dict[str, tuple[str, ...]],
    dict[str, tuple[str, ...]],
    dict[str, tuple[str, ...]],
]:
    """
    Move the input block by one unit towards the input direction.

    Parameters
    ----------
    rsc_block: RotatedSurfaceCode
        The rotated surface code block to be moved.
    direction: Direction | DiagonalDirection
        The direction towards which the block is moved.
    debug_mode: bool
        If True, skip validation when creating the new block.

    Returns
    -------
    RotatedSurfaceCode:
        The moved rotated surface code block.
    dict[str, tuple[str, ...]]:
        A dictionary mapping old stabilizer UUIDs to new stabilizer UUIDs.
    dict[str, tuple[str, ...]]:
        A dictionary mapping old logical X operator UUIDs to new logical X operator
        UUIDs.
    dict[str, tuple[str, ...]]:
        A dictionary mapping old logical Z operator UUIDs to new logical Z operator
        UUIDs.
    """
    new_stabilizers = [
        Stabilizer(
            pauli=stab.pauli,
            data_qubits=update_qubit_coords(stab.data_qubits, direction),
            ancilla_qubits=update_qubit_coords(stab.ancilla_qubits, direction),
        )
        for stab in rsc_block.stabilizers
    ]
    new_logical_x_operators = [
        PauliOperator(
            log_x.pauli,
            data_qubits=update_qubit_coords(log_x.data_qubits, direction),
        )
        for log_x in rsc_block.logical_x_operators
    ]
    new_logical_z_operators = [
        PauliOperator(
            log_z.pauli,
            data_qubits=update_qubit_coords(log_z.data_qubits, direction),
        )
        for log_z in rsc_block.logical_z_operators
    ]

    # Construct the new stabilizer to circuit map by replacing old stabilizer uuids
    # with the new ones
    stabs_map_old_to_new_uuid = {
        stab.uuid: new_stabilizers[i].uuid
        for i, stab in enumerate(rsc_block.stabilizers)
    }
    new_stabilizer_to_circuit = {
        stabs_map_old_to_new_uuid[old_stab_uuid]: circ_uuid
        for old_stab_uuid, circ_uuid in rsc_block.stabilizer_to_circuit.items()
    }

    new_block = RotatedSurfaceCode(
        unique_label=rsc_block.unique_label,
        stabilizers=new_stabilizers,
        logical_x_operators=new_logical_x_operators,
        logical_z_operators=new_logical_z_operators,
        syndrome_circuits=rsc_block.syndrome_circuits,
        stabilizer_to_circuit=new_stabilizer_to_circuit,
        skip_validation=not debug_mode,
    )

    # Create evolution maps
    stab_evolution = {
        new_stab.uuid: (stab.uuid,)
        for stab, new_stab in zip(rsc_block.stabilizers, new_stabilizers)
    }
    log_x_evolution = {
        new_log_x.uuid: (log_x.uuid,)
        for log_x, new_log_x in zip(
            rsc_block.logical_x_operators, new_logical_x_operators
        )
    }
    log_z_evolution = {
        new_log_z.uuid: (log_z.uuid,)
        for log_z, new_log_z in zip(
            rsc_block.logical_z_operators, new_logical_z_operators
        )
    }

    return new_block, stab_evolution, log_x_evolution, log_z_evolution


class DetailedSchedule(Enum):
    """
    Assumptions made (The 4-body stabilizer looks like this):
    1 -- 2
    |    |
    3 -- 4

    For N-type schedule, if the first qubit is 1. (i.e. N1)
    The schedule will be: 1 -> 3 -> 2 -> 4

    For Z-type schedule, if the first qubit is 4. (i.e. Z4)
    The schedule will be: 4 -> 3 -> 2 -> 1

    The key of the Enum is the schedule type and first qubit in the schedule.
    (i.e. N1, N2, ...)
    The value is a tuple of qubit indices in the order of the schedule.
    (i.e. [1, 3, 2, 4]: tuple[int, int, int, int])

    NOTE: If the Block was created by the RotatedSurfaceCode.create() method, the
    schedules will be N2 or Z2, for the X(Z) or Z(X) Stabilizer.

    Code example:
    ```python
    schedule = DetailedSchedule.N1
    type(schedule)  # <enum 'DetailedSchedule'>
    type(schedule.value)  # <class 'tuple'>
    ```
    """

    # N-type schedules
    N1 = (1, 3, 2, 4)  # 1 -> 3 -> 2 -> 4
    N2 = (2, 4, 1, 3)  # 2 -> 4 -> 1 -> 3
    N3 = (3, 1, 4, 2)  # 3 -> 1 -> 4 -> 2
    N4 = (4, 2, 3, 1)  # 4 -> 2 -> 3 -> 1

    # Z-type schedules
    Z1 = (1, 2, 3, 4)  # 1 -> 2 -> 3 -> 4
    Z2 = (2, 1, 4, 3)  # 2 -> 1 -> 4 -> 3
    Z3 = (3, 4, 1, 2)  # 3 -> 4 -> 1 -> 2
    Z4 = (4, 3, 2, 1)  # 4 -> 3 -> 2 -> 1

    @classmethod
    def from_schedule_and_direction(  # pylint: disable=too-many-return-statements
        cls, schedule: FourBodySchedule, starting_qubit_direction: DiagonalDirection
    ) -> DetailedSchedule:
        """
        Create a DetailedSchedule from a FourBodySchedule and a starting qubit
        direction.
        """
        match (schedule, starting_qubit_direction):
            case (FourBodySchedule.N, DiagonalDirection.TOP_LEFT):
                return cls.N1
            case (FourBodySchedule.N, DiagonalDirection.TOP_RIGHT):
                return cls.N2
            case (FourBodySchedule.N, DiagonalDirection.BOTTOM_LEFT):
                return cls.N3
            case (FourBodySchedule.N, DiagonalDirection.BOTTOM_RIGHT):
                return cls.N4
            case (FourBodySchedule.Z, DiagonalDirection.TOP_LEFT):
                return cls.Z1
            case (FourBodySchedule.Z, DiagonalDirection.TOP_RIGHT):
                return cls.Z2
            case (FourBodySchedule.Z, DiagonalDirection.BOTTOM_LEFT):
                return cls.Z3
            case (FourBodySchedule.Z, DiagonalDirection.BOTTOM_RIGHT):
                return cls.Z4
            case _:
                raise ValueError(
                    f"Invalid combination of schedule {schedule} and "
                    f"starting qubit direction {starting_qubit_direction}."
                )

    def is_N(self) -> bool:  # pylint: disable=invalid-name
        """
        Checks if the schedule is an N-type schedule.

        Returns
        -------
        bool
            True if the schedule is an N-type schedule, False otherwise.
        """
        return self.name[0] == "N"

    def is_Z(self) -> bool:  # pylint: disable=invalid-name
        """
        Checks if the schedule is a Z-type schedule.

        Returns
        -------
        bool
            True if the schedule is a Z-type schedule, False otherwise.
        """
        return self.name[0] == "Z"

    def invert_vertically(self) -> DetailedSchedule:
        """
        Inverts the schedule vertically. For example, N1 = (1, 3, 2, 4) is transformed
        into N2 = (2, 4, 1, 3).

        Returns
        -------
        DetailedSchedule
            The inverted schedule.
        """
        v_inversion_dict = {
            DetailedSchedule.N1: DetailedSchedule.N2,
            DetailedSchedule.N2: DetailedSchedule.N1,
            DetailedSchedule.N3: DetailedSchedule.N4,
            DetailedSchedule.N4: DetailedSchedule.N3,
            DetailedSchedule.Z1: DetailedSchedule.Z2,
            DetailedSchedule.Z2: DetailedSchedule.Z1,
            DetailedSchedule.Z3: DetailedSchedule.Z4,
            DetailedSchedule.Z4: DetailedSchedule.Z3,
        }
        return v_inversion_dict[self]

    def rotate_ccw_90(self) -> DetailedSchedule:
        """
        Rotates the schedule counter-clockwise by 90 degrees.
        For example, N1 = (1, 3, 2, 4) is transformed into Z3 = (3, 4, 1, 2).

        Returns
        -------
        DetailedSchedule
            The rotated schedule.
        """
        ccw_rotation_dict = {
            # First length 4 cycle
            DetailedSchedule.N1: DetailedSchedule.Z3,
            DetailedSchedule.Z3: DetailedSchedule.N4,
            DetailedSchedule.N4: DetailedSchedule.Z2,
            DetailedSchedule.Z2: DetailedSchedule.N1,
            # Second length 4 cycle
            DetailedSchedule.Z1: DetailedSchedule.N3,
            DetailedSchedule.N3: DetailedSchedule.Z4,
            DetailedSchedule.Z4: DetailedSchedule.N2,
            DetailedSchedule.N2: DetailedSchedule.Z1,
        }
        return ccw_rotation_dict[self]

    def get_stabilizer_qubits(
        self, stab: Stabilizer, boundary_direction: Direction = None
    ) -> list[tuple[int, int, int]]:
        """For a specific stabilizer and a specific schedule, return the qubits in the
        order they should be entangled. The qubits are returned in the order of the
        schedule. The qubits that are not in the schedule are set to None.

        Parameters
        ----------
        stab: Stabilizer
            The stabilizer for which the qubits are to be determined.
        boundary_direction: Direction
            The direction of the boundary. This is required for weight-2 stabilizers.

        Returns
        -------
        list[tuple[int, int, int]]
            The qubits in the order they should be entangled. The qubits that are not
            in the schedule are set to None.
        """

        # Find the qubits that correspond to the indices in the schedule
        index_to_qubit = {
            1: max(stab.data_qubits, key=lambda x: -x[0] - x[1]),
            2: max(stab.data_qubits, key=lambda x: x[0] - x[1]),
            3: max(stab.data_qubits, key=lambda x: -x[0] + x[1]),
            4: max(stab.data_qubits, key=lambda x: x[0] + x[1]),
        }

        # If the stabilizer is weight-2, we need to set the qubits that are not in the
        # schedule to None
        qubit_idxs_to_set_none = []
        if len(stab.data_qubits) == 2:
            match boundary_direction:
                case Direction.TOP:
                    qubit_idxs_to_set_none = [1, 2]
                case Direction.RIGHT:
                    qubit_idxs_to_set_none = [2, 4]
                case Direction.LEFT:
                    qubit_idxs_to_set_none = [1, 3]
                case Direction.BOTTOM:
                    qubit_idxs_to_set_none = [3, 4]
                case _:
                    raise ValueError("The boundary direction must be specified.")

        # Evaluate the qubits and pad with None if necessary
        qubits = [
            index_to_qubit[idx] if idx not in qubit_idxs_to_set_none else None
            for idx in self.value
        ]

        return qubits


def find_detailed_schedules(
    rsc_block: RotatedSurfaceCode, starting_direction: DiagonalDirection
) -> dict[Stabilizer, DetailedSchedule]:
    """
    Find the detailed schedules for each stabilizer in the input block. The detailed
    schedule is determined by the schedule type (N or Z) and the starting qubit
    direction.
    We can expand it by refactoring the code to distinguish between triangle and
    non-triangle stabilizers with a function.

    Parameters
    ----------
    rsc_block: RotatedSurfaceCode
        The rotated surface code block.
    starting_direction: DiagonalDirection
        The diagonal direction of the starting data qubit for each syndrome extraction
        circuit.
    Returns
    -------
    dict[Stabilizer, DetailedSchedule]:
        A dictionary mapping each stabilizer to its corresponding detailed schedule.
    """
    (
        non_triangle_x_schedule,
        non_triangle_z_schedule,
        triangle_x_schedule,
        triangle_z_schedule,
    ) = find_schedules(rsc_block)

    is_horizontal = rsc_block.is_horizontal
    config, pivot_corners = rsc_block.config_and_pivot_corners

    detailed_schedules_map = {}
    for stab in rsc_block.stabilizers:
        pauli_type = stab.pauli_type
        is_in_triangular, _ = find_stabilizer_position(
            rsc_block, config, pivot_corners, stab, is_horizontal
        )
        if is_in_triangular:
            schedule = triangle_x_schedule if pauli_type == "X" else triangle_z_schedule
        else:
            schedule = (
                non_triangle_x_schedule
                if pauli_type == "X"
                else non_triangle_z_schedule
            )
        detailed_schedule = DetailedSchedule.from_schedule_and_direction(
            schedule, starting_direction
        )
        detailed_schedules_map.update({stab: detailed_schedule})

    return detailed_schedules_map


def find_relative_diagonal_direction(
    from_qubit: tuple[int, ...], to_qubit: tuple[int, ...]
) -> DiagonalDirection:
    """
    Find the relative diagonal direction from qubit_1 to qubit_2.
    If two qubits have the same horizontal coordinate, the horizontal direction is
    set to Direction.LEFT by default.
    If two qubits have the same vertical coordinate, the vertical direction is
    set to Direction.TOP by default.

    Parameters
    ----------
    from_qubit: tuple[int, ...]
        The starting qubit
    to_qubit: tuple[int, ...]
        The final qubit
    """
    hor_direction = Direction.RIGHT if to_qubit[0] > from_qubit[0] else Direction.LEFT
    vert_direction = Direction.BOTTOM if to_qubit[1] > from_qubit[1] else Direction.TOP
    diagonal_direction = DiagonalDirection.from_directions(
        (hor_direction, vert_direction)
    )
    return diagonal_direction



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from functools import cached_property, reduce
from pydantic.dataclasses import dataclass
from pydantic import field_validator, model_validator


import numpy as np

from .block import Block
from .utilities.pauli_binary_vector_rep import SignedPauliOp
from .utilities.stab_array import StabArray, find_destabarray
from .utilities.validation_tools import dataclass_config


@dataclass(config=dataclass_config)
class LogicalState:
    """A logical state defined by its logical operators that stabilize it. The state is
    the statevector that is simultaneously the +1 eigenvector of all the logical
    operators stabilizing it. The operators are represented as sparse Pauli operators
    on the logical level.

    Parameters
    ----------
    sparse_logical_paulistrings : tuple[str, ...]
        A tuple of strings representing the logical state stabilizers as sparse Pauli
        operators expressed on the logical level. The strings should be in the format of
        a sparse Pauli operator, e.g. "+Z1Y3".
    """

    sparse_logical_paulistrings: tuple[str, ...]

    @classmethod
    def from_stabarray(
        cls,
        stabarray: StabArray,
    ) -> LogicalState:
        """
        Create a LogicalState from a StabArray. The StabArray should be on the logical
        level, meaning that every column corresponds to a logical qubit.
        """
        if not isinstance(stabarray, StabArray):
            raise TypeError(
                "The input should be a StabArray object representing the logical "
                "operators."
            )
        sparse_logical_paulistrings = tuple(
            signed_pauli_op.as_sparse_string() for signed_pauli_op in stabarray
        )
        return LogicalState(sparse_logical_paulistrings=sparse_logical_paulistrings)

    @field_validator("sparse_logical_paulistrings", mode="before")
    @classmethod
    def _cast_to_uppercase(cls, value: tuple[str]):
        """
        Cast the sparse_logical_operators to uppercase.
        """
        if isinstance(value, str):
            value = (value,)
        return tuple(op.upper() for op in value)

    @field_validator("sparse_logical_paulistrings", mode="before")
    @classmethod
    def cast_str_to_tuple(cls, value):
        """Cast the sparse_logical_paulistrings to a tuple if it is a single string."""
        if isinstance(value, str):
            return (value,)
        return value

    @field_validator("sparse_logical_paulistrings", mode="after")
    @classmethod
    def _validate_correct_format(cls, value: tuple[str]):
        """
        Check if the sparse_logical_paulistrings are in the correct format.
        """
        # Check that the sparse_logical_operators are in the correct format by
        # attempting to initialize SignedPauliOps
        signed_pauli_ops = [SignedPauliOp.from_sparse_string(op) for op in value]

        # Find the maximum number of logical qubits required and check that it matches
        # the number of sparse Pauli strings
        max_qubits = max(op.nqubits for op in signed_pauli_ops)

        if max_qubits != len(value):
            raise ValueError(
                f"The number of sparse stabilizers ({len(value)}) does not "
                "match the maximum logical qubits required by the operators "
                f"({max_qubits})."
            )
        return value

    @model_validator(mode="after")
    def _validate_irreducibility(self):
        """
        Check that the the set of logical paulistrings is not irreducible.
        """
        if not self.stabarray.is_irreducible:
            raise ValueError("The set of logical paulistrings is not irreducible.")
        return self

    def __repr__(self) -> str:
        return f"LogicalState({self.sparse_logical_paulistrings})"

    @property
    def n_logical_qubits(self) -> int:
        """Return the number of logical qubits."""
        return len(self.sparse_logical_paulistrings)

    @cached_property
    def stabarray(self) -> StabArray:
        """
        Return the StabArray representation of the logical operator set that stabilizes
        the state. The representation is on the logical level which means that every
        column in the StabArray corresponds to a logical qubit rather than a
        data/physical qubit.
        """
        return StabArray.from_signed_pauli_ops(
            [
                SignedPauliOp.from_sparse_string(op, nqubits=self.n_logical_qubits)
                for op in self.sparse_logical_paulistrings
            ]
        )

    @cached_property
    def destabarray(self) -> StabArray:
        """
        Return the StabArray representation of the destabilizer array corresponding to
        the stabarray property.
        """
        return find_destabarray(self.stabarray)

    @classmethod
    def convert_logical_to_base_representation(
        cls, block: Block, logical_stabarray: StabArray
    ) -> StabArray:
        """
        Convert a logical StabArray representation to the base StabArray representation
        of a Block object. If the Block object describes a code with n data qubits
        encoding k logical qubits, then the logical_stabarray should be
        indexing the logical qubits from 0 to k-1. The base representation that will be
        returned will be indexing the data qubits from 0 to n-1.

        Parameters
        ----------
        block : Block
            The Block object to use to convert the logical StabArray.
        logical_stabarray : StabArray
            The logical StabArray representation to convert.

        Returns
        -------
        StabArray
            The base StabArray representation of the input logical StabArray.
        """
        if block.n_logical_qubits != logical_stabarray.nqubits:
            raise ValueError(
                f"The number of logical qubits {block.n_logical_qubits} in the Block "
                "does not match the number of logical qubits "
                f"{logical_stabarray.nqubits} in the logical StabArray."
            )

        operators_in_base_repr = [
            cls.convert_logical_pauli_op_to_base_representation(
                log_op, block.x_log_stabarray, block.z_log_stabarray
            )
            for log_op in logical_stabarray
        ]

        return StabArray.from_signed_pauli_ops(operators_in_base_repr)

    @staticmethod
    def convert_logical_pauli_op_to_base_representation(
        log_op: SignedPauliOp, x_log_stabarray: StabArray, z_log_stabarray: StabArray
    ) -> SignedPauliOp:
        """
        Convert a logical Pauli operator to its base representation. The base
        representation is defined as the representation of the operator on the data
        qubits of a Block object. The logical Pauli operator is expected to be
        represented as a SignedPauliOp object, and the x_log_stabarray and
        z_log_stabarray are the StabArray representations of the logical X and Z
        operators, respectively.

        Parameters
        ----------
        log_op : SignedPauliOp
            The logical Pauli operator to convert.
        x_log_stabarray : StabArray
            The StabArray representation of the logical X operators.
        z_log_stabarray : StabArray
            The StabArray representation of the logical Z operators.

        Returns
        -------
        SignedPauliOp
            The base representation of the logical Pauli operator.
        """
        n_data_qubits = x_log_stabarray.nqubits
        where_z = np.where(log_op.z == 1)[0]
        where_x = np.where(log_op.x == 1)[0]

        # for every logical operator find the corresponding Z, X, and Y operators
        y_op_indexes = np.intersect1d(where_z, where_x)
        z_op_indexes = np.setdiff1d(where_z, y_op_indexes)
        x_op_indexes = np.setdiff1d(where_x, y_op_indexes)

        # Construct the Y operators from the Z and X operators.
        # Note that Y = i * X * Z, so we need to have the appropriate order of
        # multiplication to get the correct sign.
        y_operators = [
            x_log_stabarray[y_idx].multiply_with_anticommuting_operator(
                z_log_stabarray[y_idx]
            )
            for y_idx in y_op_indexes
        ]

        # Multiply all the Z, X, and Y operators together to get the logical
        # operator in the base representation.
        operator = reduce(
            lambda a, b: a * b,
            [z_log_stabarray[z_idx] for z_idx in z_op_indexes]
            + [x_log_stabarray[x_idx] for x_idx in x_op_indexes]
            + y_operators,
            SignedPauliOp.identity(n_data_qubits, negative=log_op.sign),
        )

        return operator

    def get_tableau(self, block: Block) -> np.ndarray:
        """Given a Block, return the tableau of the logical state.

        Parameters
        ----------
        block : Block
            The Block object to use to generate the tableau.

        Returns
        -------
        np.ndarray
            The tableau of the logical state.
        """
        if block.n_logical_qubits != self.n_logical_qubits:
            raise ValueError(
                "The number of logical qubits in the Block does not match "
                "the number of logical qubits in the LogicalState."
            )
        # get the base representation of the logical operators
        stab_log_ops = self.convert_logical_to_base_representation(
            block, self.stabarray
        )
        # get the base representation of the destabilizers of the logical operators
        destab_log_ops = self.convert_logical_to_base_representation(
            block, self.destabarray
        )
        # return the tableau by concatenating appropriately the arrays
        return np.vstack(
            (
                block.destabarray.array,
                destab_log_ops.array,
                block.reduced_stabarray.array,
                stab_log_ops.array,
            )
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Eka
from loom.interpreter.applicator import CodeApplicator

from .grow import grow
from .shrink import shrink
from .merge import merge
from .split import split
from .auxcnot import auxcnot

# from .transversalhadamard import transversalhadamard
from .move_block import move_block
from .rotate_block import rotate_block
from .transversalhadamard import transversalhadamard
from ..code_factory import RotatedSurfaceCode
from .logical_phase_via_ywall import logical_phase_via_ywall
from .state_injection import state_injection


class RotatedSurfaceCodeApplicator(  # pylint: disable=too-few-public-methods
    CodeApplicator
):
    """
    Contains the implementation logic for each operation, for the Rotated Surface Code.
    """

    def __init__(
        self,
        eka: Eka,
    ):
        # Ensure that all blocks are typed RotatedSurfaceCode
        if any(not isinstance(block, RotatedSurfaceCode) for block in eka.blocks):
            raise ValueError("All blocks must be of type RotatedSurfaceCode.")
        super().__init__(eka)
        # Add the extra operations that are supported by the Rotated Surface Code
        self.supported_operations |= {
            "Grow": grow,
            "Shrink": shrink,
            "Merge": merge,
            "Split": split,
            "AuxCNOT": auxcnot,
            "TransversalHadamard": transversalhadamard,
            "MoveBlock": move_block,
            "LogicalPhaseViaYwall": logical_phase_via_ywall,
            "StateInjection": state_injection,
            "RotateBlock": rotate_block,
        }
